# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2019 Maximilian Stahlberg
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Console output helpers related to formatting."""

from __future__ import print_function

import re
import sys
from contextlib import contextmanager
from math import ceil, floor

from . import glyphs
from .apidoc import api_end, api_start

_API_START = api_start(globals())
# -------------------------------


HEADER_WIDTH = 35  #: Character length of headers and footers printed by PICOS.


def print_header(title, subtitle=None, symbol="-", width=HEADER_WIDTH):
    """Print a header line."""
    w = "{:d}".format(width)

    print(("{0}\n{1:^"+w+"}\n{2}{0}").format(symbol * width, title,
        ("{:^"+w+"}\n").format("{}".format(subtitle)) if subtitle else ""))
    sys.stdout.flush()


def print_footer(caption, symbol="=", width=HEADER_WIDTH):
    """Print a footer line."""
    middle = "[ {} ]".format(caption)

    if width < len(middle) + 2:
        footer = symbol * width
    else:
        s = (width - len(middle))
        l = int(floor(s / 2.0))
        r = int(ceil(s / 2.0))
        footer = symbol * l + middle + symbol * r

    print(footer)
    sys.stdout.flush()


@contextmanager
def box(title, caption, subtitle=None, symbol="-", width=HEADER_WIDTH,
        show=True):
    """Print both a header above and a footer below the context.

    :param str title: The (long) main title printed at the top.
    :param str caption: The (short) caption printed at the bottom.
    :param str subtitle: The (long) subtitle printed at the top.
    :param str symbol: A single character used to draw lines.
    :param int width: The width of the box.
    :param bool show: Whether anything should be printed at all.
    """
    if show:
        print_header(title, subtitle, symbol, width)
        yield
        print_footer(caption, symbol, width)
    else:
        yield


@contextmanager
def picos_box(show=True):
    """Print a PICOS header above and a PICOS footer below the context."""
    from .__init__ import __version__ as picosVer
    with box("PICOS {}".format(picosVer), "PICOS", symbol="=", show=show):
        yield


@contextmanager
def solver_box(longName, shortName, subSolver=None, show=True):
    """Print a solver header above and a solver footer below the context."""
    subtitle = "via {}".format(subSolver) if subSolver else None
    with box(longName, shortName, subtitle, show=show):
        yield


def doc_cat(docstring, append):
    """Append to a docstring."""
    if not docstring:
        raise ValueError("Empty base docstring.")

    if not append:
        return docstring

    lines = [line for line in docstring.splitlines() if line.strip()]
    i = len(lines[1]) - len(lines[1].lstrip()) if len(lines) > 1 else 0

    append = "\n".join(" "*i + line for line in append.splitlines())

    return docstring.rstrip() + "\n\n" + append


def detect_range(sequence, asQuadruple=False, asStringTemplate=False,
        shortString=False):
    """Return a Python range mirroring the given integer sequence.

    :param sequence: An integer sequence that can be mirrored by a Python range.
    :param bool asQuadruple: Whether to return a quadruple with factor, inner
        shift, outer shift, and length, formally ``(a, i, o, n)`` such that
        ``[a*(x+i)+o for x in range(n)]`` mirrors the input sequence.
    :param bool asStringTemplate: Whether to return a format string that, if
        instanciated with numbers from ``0`` to ``len(sequence) - 1``, yields
        math expression strings that describe the input sequence members.
    :param bool shortString: Whether to return condensed string templates that
        are designed to be instanciated with an index character string. Requires
        asStringTemplate to be ``True``.
    :raises TypeError: If the input is not an integer sequence.
    :raises ValueError: If the input cannot be mirrored by a Python range.
    :returns: A range object, a quadruple of numbers, or a format string.

    :Example:

    >>> from picos.formatting import detect_range as dr
    >>> R = range(7,30,5)
    >>> S = list(R)
    >>> S
    [7, 12, 17, 22, 27]
    >>> # By default, returns a matching range object:
    >>> dr(S)
    range(7, 28, 5)
    >>> dr(S) == R
    True
    >>> # Sequence elements can also be decomposed w.r.t. range(len(S)):
    >>> a, i, o, n = dr(S, asQuadruple=True)
    >>> [a*(x+i)+o for x in range(n)] == S
    True
    >>> # The same decomposition can be returned in a string representation:
    >>> dr(S, asStringTemplate=True)
    '5·({} + 1) + 2'
    >>> # Short string representations are designed to accept index names:
    >>> dr(S, asStringTemplate=True, shortString=True).format("i")
    '5(i+1)+2'
    >>> dr(range(0,100,5), asStringTemplate=True, shortString=True).format("i")
    '5i'
    >>> dr(range(10,100), asStringTemplate=True, shortString=True).format("i")
    'i+10'

    :Example:

    >>> # This works with decreasing ranges as well.
    >>> R2 = range(10,4,-2)
    >>> S2 = list(R2)
    >>> S2
    [10, 8, 6]
    >>> dr(S2)
    range(10, 5, -2)
    >>> dr(S2) == R2
    True
    >>> a, i, o, n = dr(S2, asQuadruple=True)
    >>> [a*(x+i)+o for x in range(n)] == S2
    True
    >>> T = dr(S2, asStringTemplate=True, shortString=True)
    >>> [T.format(i) for i in range(len(S2))]
    ['-2(0-5)', '-2(1-5)', '-2(2-5)']
    """
    if asQuadruple and asStringTemplate:
        raise ValueError(
            "Can only return a quadruple or a string template, not both.")

    if shortString and not asStringTemplate:
        raise ValueError("Enabling 'shortString' requires 'asStringTemplate'.")

    if len(sequence) == 0:
        if asQuadruple:
            return 0, 0, 0, 0
        elif asStringTemplate:
            return ""
        else:
            return range(0)

    first  = sequence[0]
    last   = sequence[-1]
    next   = last + (1 if first <= last else -1)
    length = len(sequence)

    if not isinstance(first, int) or not isinstance(last, int):
        raise TypeError("Not an integer container.")

    # Determine potential integer step size.
    if length > 1:
        step = (last - first) / (length - 1)
    else:
        step = 1
    if int(step) != step:
        raise ValueError("Cannot be mirrored by a Python range.")
    step = int(step)

    # Determine potential range.
    range_ = range(first, next, step)
    for position, number in enumerate(range_):
        if sequence[position] != number:
            raise ValueError("Cannot be mirrored by a Python range.")

    if asQuadruple or asStringTemplate:
        # Compute inner and outer shift.
        innerShift = first // step
        outerShift = first  % step

        # Verify our finding.
        assert last // step + 1 - innerShift                 == length
        assert step*(0 + innerShift) + outerShift            == first
        assert step*((length - 1) + innerShift) + outerShift == last

        if asQuadruple:
            return step, innerShift, outerShift, length
        elif shortString:
            string = "{{}}{:+d}".format(innerShift) if innerShift else "{}"
            if step != 1 and innerShift:
                string = "{}({})".format("-" if step == -1 else step, string)
            elif step != 1:
                string = "{}{}".format("-" if step == -1 else step, string)
            string = "{}{:+d}".format(string, outerShift) \
                if outerShift else string
            # TODO: Something like the following is needed in case replacement
            #       happens for a factor in a multiplication.
            # if (innerShift and step == 1) or outerShift:
            #     string = "({})".format(string)

            return string
        else:
            glyph = glyphs.add("{}", innerShift) if innerShift else "{}"
            glyph = glyphs.mul(step, glyph) if step != 1 else glyph
            glyph = glyphs.add(glyph, outerShift) if outerShift else glyph

            return glyph
    else:
        return range_


def natsorted(strings, key=None):
    """Sort the given list of strings naturally with respect to numbers."""
    def natsplit(string):
        parts = re.split(r"([+-]?\d*\.?\d+)", string)
        split = []
        for part in parts:
            try:
                split.append(float(part))
            except ValueError:
                split.append(part)
        return split

    if key:
        return sorted(strings, key=lambda x: natsplit(key(x)))
    else:
        return sorted(strings, key=natsplit)


def parameterized_string(
        strings, replace=r"-?\d+", context=r"\W", placeholders="ijklpqr",
        fallback="?"):
    """Find a string template for the given (algebraic) strings.

    Given a list of strings with similar structure, finds a single string with
    placeholders and an expression that denotes how to instantiate the
    placeholders in order to obtain each string in the list.

    The function is designed to take a number of symbolic string representations
    of math expressions that differ only with respect to indices.

    :param list strings:
        The list of strings to compare.
    :param str replace:
        A regular expression describing the bits to replace with placeholders.
    :param str context:
        A regular expression describing context characters that need to be
        present next to the bits to be replaced with placeholders.
    :param placeholders:
        An iterable of placeholder strings. Usually a string, so that each of
        its characters becomes a placeholder.
    :param str fallback:
        A fallback placeholder string, if the given placeholders are not
        sufficient.

    :returns:
        A tuple of two strings, the first being the template string and the
        second being a description of the placeholders used.

    :Example:

    >>> from picos.formatting import parameterized_string as ps
    >>> ps(["A[{}]".format(i) for i in range(5, 31)])
    ('A[i+5]', 'i ∈ [0…25]')
    >>> ps(["A[{}]".format(i) for i in range(5, 31, 5)])
    ('A[5(i+1)]', 'i ∈ [0…5]')
    >>> S=["A[0]·B[2]·C[3]·D[5]·F[0]",
    ...    "A[1]·B[1]·C[6]·D[6]·F[0]",
    ...    "A[2]·B[0]·C[9]·D[9]·F[0]"]
    >>> ps(S)
    ('A[i]·B[-(i-2)]·C[3(i+1)]·D[j]·F[0]', '(i,j) ∈ zip([0…2],[5,6,9])')
    """
    if len(strings) == 0:
        return "", ""
    elif len(strings) == 1:
        return strings[0], ""

    for string in strings:
        if not isinstance(string, str):
            raise TypeError("First argument must be a list of strings.")

    def split_with_context(string, match, context):
        pattern = r"(?<={0}){1}(?={0})".format(context, match)
        return re.split(pattern, string)

    def findall_with_context(string, match, context):
        pattern = r"(?<={0}){1}(?={0})".format(context, match)
        return re.findall(pattern, string)

    # The skeleton of a string is the part not matched by 'replace' and
    # surrounded by 'context', and it must be the same for all strings.
    skeleton = split_with_context(strings[0], replace, context)
    for string in strings[1:]:
        if skeleton != split_with_context(string, replace, context):
            raise ValueError("Strings do not have a common skeleton.")

    # The slots are the parts that are matched by 'replace' and surrounded by
    # 'context' and should be filled  with the placeholders.
    slotToValues = []
    for string in strings:
        slotToValues.append(findall_with_context(string, replace, context))
    slotToValues = list(zip(*slotToValues))

    # Verify that slots are always surrounded by (empty) skeleton strings.
    assert len(skeleton) == len(slotToValues) + 1

    # Find slots with the same value in each string; add them to the skeleton.
    for slot in range(len(slotToValues)):
        if len(set(slotToValues[slot])) == 1:
            skeleton[slot + 1] =\
                skeleton[slot] + slotToValues[slot][0] + skeleton[slot + 1]
            skeleton[slot] = None
            slotToValues[slot] = None
    skeleton     = [s for s in skeleton     if s is not None]
    slotToValues = [v for v in slotToValues if v is not None]

    # We next build a mapping from slots to (few) placeholder indices.
    slotToIndex = {}
    nextIndex = 0

    # Find slots whose values form a range, and build string templates that lift
    # a placeholder to a formula denoting sequence elements (e.g. "i" → "2i+1").
    # All such slots share the first placeholder (with index 0).
    slotToTemplate = {}
    for slot, values in enumerate(slotToValues):
        try:
            slotToTemplate[slot] = detect_range([int(v) for v in values],
                asStringTemplate=True, shortString=True)
        except ValueError:
            pass
        else:
            slotToIndex[slot] = 0
            nextIndex = 1

    # Find slots with identical value in each string and assign them the same
    # placeholder.
    valsToIndex = {}
    for slot, values in enumerate(slotToValues):
        if slot in slotToIndex:
            # The slot holds a range.
            continue

        if values in valsToIndex:
            slotToIndex[slot] = valsToIndex[values]
        else:
            slotToIndex[slot]   = nextIndex
            valsToIndex[values] = nextIndex
            nextIndex += 1

    # Define a function that maps slots to their placeholder symbols.
    def placeholder(slot):
        index = slotToIndex[slot]
        return placeholders[index] if index < len(placeholders) else fallback

    # Assemble the string template (with values replaced by placeholders).
    template = ""
    for slot in range(len(slotToIndex)):
        if slot in slotToTemplate:
            ph = slotToTemplate[slot].format(placeholder(slot))
        else:
            ph = placeholder(slot)

        template += skeleton[slot] + ph
    template += skeleton[-1]

    # Collect the placeholdes that were used, and their domains.
    usedPHs, domains = [], []
    indices = set()
    for slot, index in slotToIndex.items():
        values = slotToValues[slot]

        if index in indices:
            continue
        else:
            indices.add(index)

        usedPHs.append(placeholder(slot))

        if slot in slotToTemplate:
            domains.append(glyphs.intrange(0, len(values) - 1))
        elif len(values) > 4:
            domains.append(glyphs.intrange(
                ",".join(values[:2]) + ",", "," + ",".join(values[-2:])))
        else:
            domains.append(glyphs.interval(",".join(values)))

    # Make sure used placeholders and their domains match in number.
    assert len(usedPHs) == len(domains)

    # Assemble occuring placeholders and ther joint domain (the data).
    if len(domains) == 0:
        data = ""
    else:
        if len(domains) == 1:
            usedPHs = usedPHs[0]
            domain  = domains[0]
        else:
            usedPHs = "({})".format(",".join(usedPHs))
            domain  = "zip({})".format(",".join(domains))

        data = glyphs.element(usedPHs, domain)

    return template, data


# --------------------------------------
__all__ = api_end(_API_START, globals())



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2018-2019 Maximilian Stahlberg
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Second order conce constraints."""

from collections import namedtuple

from .. import glyphs
from ..apidoc import api_end, api_start
from .constraint import Constraint

_API_START = api_start(globals())
# -------------------------------


class SOCConstraint(Constraint):
    """Second order (:math:`2`-norm, Lorentz) cone membership constraint."""

    def __init__(self, normedExpression, upperBound, customString=None):
        """Construct a :class:`SOCConstraint`.

        :param ~picos.expressions.AffineExpression normedExpression:
            Expression under the norm.
        :param ~picos.expressions.AffineExpression upperBound:
            Upper bound on the normed expression.
        :param str customString:
            Optional string description.
        """
        from ..expressions import AffineExpression

        assert isinstance(normedExpression, AffineExpression)
        assert isinstance(upperBound, AffineExpression)
        assert len(upperBound) == 1

        # NOTE: len(normedExpression) == 1 is allowed even though this should
        #       rather be represented as an AbsoluteValueConstraint.

        self.ne = normedExpression
        self.ub = upperBound

        super(SOCConstraint, self).__init__(
            self._get_type_term(), customString, printSize=True)

    def _get_type_term(self):
        return "SOC"

    Subtype = namedtuple("Subtype", ("argdim",))

    def _subtype(self):
        return self.Subtype(len(self.ne))

    @classmethod
    def _cost(cls, subtype):
        return subtype.argdim + 1

    def _expression_names(self):
        yield "ne"
        yield "ub"

    def _str(self):
        return glyphs.le(glyphs.norm(self.ne.string), self.ub.string)

    def _get_size(self):
        return (len(self.ne) + 1, 1)

    def _get_slack(self):
        return self.ub.value - abs(self.ne).value


# --------------------------------------
__all__ = api_end(_API_START, globals())



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2012-2017 Guillaume Sagnol
# Copyright (C) 2018-2019 Maximilian Stahlberg
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Implementation of :class:`GeometricMeanConstraint`."""

import math
from collections import namedtuple

from .. import glyphs
from ..apidoc import api_end, api_start
from .constraint import Constraint, ConstraintConversion

_API_START = api_start(globals())
# -------------------------------


class GeometricMeanConstraint(Constraint):
    """Lower bound on a geometric mean."""

    class RSOCConversion(ConstraintConversion):
        """Geometric mean to rotated second order cone constraint conversion."""

        @classmethod
        def predict(cls, subtype, options):
            """Implement :meth:`~.constraint.ConstraintConversion.predict`."""
            from ..expressions import RealVariable
            from . import RSOCConstraint

            j = subtype.argdim - 1
            k = int(math.log(j, 2))
            l = j + k - "{:b}".format(j - 2**k).count("1")

            yield ("var", RealVariable.make_var_type(dim=1, bnd=0), l - 1)
            yield ("con", RSOCConstraint.make_type(argdim=1), l)

        # TODO: Refactor to be human readable.
        @classmethod
        def convert(cls, con, options):
            """Implement :meth:`~.constraint.ConstraintConversion.convert`."""
            from ..expressions.algebra import rsoc
            from ..modeling import Problem

            geoMean    = con.geoMean
            lowerBound = con.lowerBound

            P     = Problem()
            x     = geoMean.x
            m     = len(x)
            lm    = [[i] for i in range(m - 1, -1, -1)]
            K     = []
            depth = 0
            u     = {}

            while len(lm) > 1:
                depth += 1
                nlm = []
                while lm:
                    i1 = lm.pop()[-1]
                    if lm:
                        i2 = lm.pop()[0]
                    else:
                        i2 = 'x'
                    nlm.insert(0, (i2, i1))
                    k = str(depth) + ':' + str(i1) + '-' + str(i2)
                    K.append(k)
                    u[k] = P.add_variable('__u[' + k + ']', 1)
                lm = nlm

            root = K[-1]
            maxd = int(K[-1].split(':')[0])
            P.remove_variable(u[root].name)  # TODO: Update for new expressions.
            u[root] = lowerBound

            for k in K:
                i1 = int(k.split('-')[0].split(':')[1])
                i2 = k.split('-')[1]
                if i2 != 'x':
                    i2 = int(i2)
                if k[:2] == '1:':
                    if i2 != 'x':
                        P.add_constraint((x[i1] & x[i2] & u[k]) << rsoc())
                    else:
                        P.add_constraint((x[i1] & lowerBound & u[k]) << rsoc())
                else:
                    d = int(k.split(':')[0])
                    if i2 == 'x' and d < maxd:
                        k2pot = [ki for ki in K
                            if ki.startswith(str(d - 1) + ':')
                            and int(ki.split(':')[1].split('-')[0]) >= i1]
                        k1 = k2pot[0]
                        if len(k2pot) == 2:
                            k2 = k2pot[1]
                            P.add_constraint((u[k1] & u[k2] & u[k]) << rsoc())
                        else:
                            P.add_constraint(
                                (u[k1] & lowerBound & u[k]) << rsoc())
                    else:
                        k1 = [ki for ki in K
                            if ki.startswith(str(d - 1) + ':' + str(i1))][0]
                        k2 = [ki for ki in K if ki.startswith(str(d - 1) + ':')
                            and ki.endswith('-' + str(i2))][0]
                        P.add_constraint((u[k1] & u[k2] & u[k]) << rsoc())

            return P

    def __init__(self, geoMean, lowerBound):
        """Construct a :class:`GeometricMeanConstraint`.

        :param ~picos.expressions.GeometricMean lhs:
            Constrained expression.
        :param ~picos.expressions.AffineExpression lowerBound:
            Lower bound on the expression.
        """
        from ..expressions import AffineExpression, GeometricMean

        assert isinstance(geoMean, GeometricMean)
        assert isinstance(lowerBound, AffineExpression)
        assert len(lowerBound) == 1

        self.geoMean    = geoMean
        self.lowerBound = lowerBound

        super(GeometricMeanConstraint, self).__init__(geoMean._typeStr)

    Subtype = namedtuple("Subtype", ("argdim",))

    def _subtype(self):
        return self.Subtype(len(self.geoMean.x))

    @classmethod
    def _cost(cls, subtype):
        return subtype.argdim + 1

    def _expression_names(self):
        yield "geoMean"
        yield "lowerBound"

    def _str(self):
        return glyphs.ge(self.geoMean.string, self.lowerBound.string)

    def _get_slack(self):
        return self.geoMean.value - self.lowerBound.value


# --------------------------------------
__all__ = api_end(_API_START, globals())



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2019 Maximilian Stahlberg
# Based on the original picos.expressions module by Guillaume Sagnol.
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

r"""Implement special matrix vectorization formats.

These formats are used to efficiently store structured variable types such as
symmetric matrix variables in the form of real variable vectors.
"""

import random
from copy import copy

import cvxopt

from .. import glyphs
from ..apidoc import api_end, api_start
from ..compat import ABC, abstractmethod
from .data import cvxopt_equals, cvxopt_hcat, cvxopt_vcat, load_shape

_API_START = api_start(globals())
# -------------------------------


#: Number of instances to cache per vectorization format.
CACHE_SIZE = 100

#: Number of cached instances to drop at random when the cache is full.
CACHE_BULK_REMOVE = CACHE_SIZE // 4


class BaseVectorization(ABC, object):
    """Abstract base class for special matrix vectorization formats.

    Subclass instances are cached: If multiple instances of the same
    vectorization format and concerning matrices of the same shape are requested
    successively, then the instance created to serve the first request is
    retrieved from a cache on successive requests. The module attributes
    :data:`CACHE_SIZE` and :data:`CACHE_BULK_REMOVE` control the size of the
    cache for each vectorization format.

    .. warning::

        Due to how caching is implemented, derived classes may not inherit from
        each other but only from :class:`BaseVectorization` directly!
    """

    def __new__(cls, shape):
        """Lookup or create a vectorization format for a fixed matrix shape."""
        shape = load_shape(shape, squareMatrix=cls._square_input())

        if not hasattr(cls, "_cache"):
            cls._cache = {}

        if shape not in cls._cache:
            if len(cls._cache) >= CACHE_SIZE:
                for remove in random.sample(
                        list(cls._cache.keys()), CACHE_BULK_REMOVE):
                    cls._cache.pop(remove)

            cls._cache[shape] = object.__new__(cls)

        return cls._cache[shape]

    def __init__(self, shape):
        """Initialize a vectorization format for a fixed matrix shape."""
        self._shape = load_shape(shape, squareMatrix=self._square_input())
        self._special2full = self._make_special_to_full()
        self._full2special = self._make_full_to_special()

    def __len__(self):
        return self._shape[0] * self._shape[1]

    @property
    def shape(self):
        """The shape of matrices being vectorized."""
        return self._shape

    @property
    def dim(self):
        """The length of the vectorization.

        This corresponds to the dimension of a matrix variable being vectorized.
        """
        return self._special2full.size[1]

    @classmethod
    @abstractmethod
    def _square_input(cls):
        """Whether input matrices must be square."""
        pass

    @abstractmethod
    def _validate_matrix(self, matrix):
        """Raise an exception if the given matrix cannot be vectorized."""
        if not isinstance(matrix, (cvxopt.matrix, cvxopt.spmatrix)):
            raise TypeError("May only vectorize CVXOPT matrix types.")

        if matrix.size != self._shape:
            raise TypeError("Cannot vectorize a matrix of shape {} according "
                "to a vectorization recipe for {} matrices."
                .format(glyphs.shape(matrix.size), glyphs.shape(self._shape)))

    def _ensure_real(self, matrix):
        """Raise a :exc:`TypeError` if the given matrix is not real."""
        if matrix.typecode == "z":
            raise TypeError(
                "The vectorization format does not support complex input.")

    def _validate_vector(self, vector):
        """Raise an exception if the given vector cannot be devectorized."""
        if not isinstance(vector, (cvxopt.matrix, cvxopt.spmatrix)):
            raise TypeError("May only devectorize CVXOPT matrix types.")

        if vector.typecode == "z":
            raise TypeError("Cannot devectorize a complex vector: All "
                "vectorizations are expected to be real.")

        if vector.size != (self.dim, 1):
            raise TypeError(
                "Invalid shape of vectorized data: Expected {} but got {}."
                .format(glyphs.shape((self.dim, 1), vector.size)))

    @abstractmethod
    def _make_special_to_full(self):
        """Return a mapping from the special to the full vectorization."""
        pass

    @abstractmethod
    def _make_full_to_special(self):
        """Return a mapping from the full to the special vectorization.

        Returns :obj:`None` if the input matrix is complex. Then,
        :meth:`vectorize` needs to be overridden.
        """
        pass

    @property
    def identity(self):
        """A linear mapping from the special to the full vectorization.

        The term *identity* comes from the fact that these matrices are used
        as the coefficients that map the internal (vectorized) representation of
        a :class:`~.variables.BaseVariable` object to the
        :class:`~.exp_affine.ComplexAffineExpression` that it represents.
        """
        return self._special2full

    def vectorize(self, matrix):
        """Given a matrix, return its special vectorization.

        :raises TypeError: If the input isn't a CVXOPT matrix or does not have
            the expected numeric type or shape.
        :raises ValueError: If the matrix does not have the expected structure.
        """
        self._validate_matrix(matrix)
        return self._full2special*matrix[:]

    def devectorize(self, vector):
        """Given a special vectorization, return the corresponding matrix.

        :raises TypeError: If the input isn't a CVXOPT column vector or does not
            have the expected numeric type or length.
        """
        self._validate_vector(vector)
        M = self._special2full*vector
        M.size = self._shape
        return M


class FullVectorization(BaseVectorization):
    """A basic column-major matrix vectorization."""

    @classmethod
    def _square_input(cls):
        return False

    def _validate_matrix(self, matrix):
        BaseVectorization._validate_matrix(self, matrix)
        self._ensure_real(matrix)

    def _make_full_to_special(self):
        n = len(self)
        return cvxopt.spmatrix([1]*n, range(n), range(n), tc="d")

    def vectorize(self, matrix):
        """Override :meth:`BaseVectorization.vectorize` for speed reasons."""
        self._validate_matrix(matrix)

        if matrix.typecode == "d":
            return matrix[:]
        else:
            assert matrix.typecode == "i"
            return self._full2special*matrix[:]

    def _make_special_to_full(self):
        # Not actually used, see devectorize.
        return self._make_full_to_special()

    def devectorize(self, vector):
        """Override :meth:`BaseVectorization.devectorize` for speed reasons."""
        self._validate_vector(vector)
        M = copy(vector)
        M.size = self._shape
        return M


class ComplexVectorization(BaseVectorization):
    """An isometric vectorization that stacks real and imaginary parts."""

    @classmethod
    def _square_input(cls):
        return False

    def _validate_matrix(self, matrix):
        BaseVectorization._validate_matrix(self, matrix)

    def _make_full_to_special(self):
        # No such matrix exists as the full vectorization is complex.
        return None

    def vectorize(self, matrix):
        """Override :meth:`BaseVectorization.vectorize`.

        This is necessary because extracting the real and the imaginary part
        cannot be done with a linear transformation matrix as is done for other
        vectorization formats.
        """
        self._validate_matrix(matrix)
        fullVectorization = matrix[:]
        return cvxopt_vcat([fullVectorization.real(), fullVectorization.imag()])

    def _make_special_to_full(self):
        cRank = len(self)  # Rank on the complex field.
        rRank = 2*cRank    # Rank on the real field.
        return cvxopt.spmatrix([1]*cRank + [1j]*cRank,
            list(range(cRank))*2, range(rRank), tc="z")


class SymmetricVectorization(BaseVectorization):
    """An isometric symmetric matrix vectorization.

    See [svec]_ for the precise vectorization used.

    .. [svec] Dattorro, J. (2018). Isomorphism of symmetric matrix subspace.
        In *Convex Optimization & Euclidean Distance Geometry (2nd ed.)*
        (pp. 47f.). California, Meboo Publishing USA. Retrieved from
        `<https://meboo.convexoptimization.com/Meboo.html>`_.
    """

    @classmethod
    def _square_input(cls):
        return True

    def _validate_matrix(self, matrix):
        BaseVectorization._validate_matrix(self, matrix)
        self._ensure_real(matrix)

        if not cvxopt_equals(matrix, matrix.T):
            raise ValueError("The given matrix is not numerically symmetric.")

    def _make_full_to_special(self):
        n = self._shape[0]
        m = n*(n + 1) // 2
        I, J, V = [], [], []
        for i in range(m):
            c = int((1 + 8*i)**0.5 - 1) // 2
            r = i - c*(c + 1) // 2

            # Version 1: Average elements in lower and upper triangular part.
            #            This could work around noisy input data.
            if c == r:
                I.append(i)
                J.append(c*n + r)
                V.append(1)
            else:
                I.extend([i, i])
                J.extend([c*n + r, r*n + c])
                V.extend([2**0.5 / 2, 2**0.5 / 2])

            # Version 2: Ignore elements below the main diagonal.
            #            This is the faster approach.
            # I.append(i)
            # J.append(c*n + r)
            # V.append(1 if c == r else 2**0.5)

        return cvxopt.spmatrix(V, I, J, (m, n**2), tc="d")

    def _make_special_to_full(self):
        n = self._shape[0]
        I, J, V = range(n**2), [], []
        for i in I:
            c, r = divmod(i, n)
            if c < r:  # Entries below the diagonal are infered.
                c, r = r, c
            J.append(c*(c + 1) // 2 + r)
            V.append(1 if c == r else 1 / 2**0.5)
        return cvxopt.spmatrix(V, I, J, (n**2, n*(n + 1) // 2), tc="d")


class SkewSymmetricVectorization(BaseVectorization):
    """An isometric skew-symmetric matrix vectorization."""

    @classmethod
    def _square_input(cls):
        return True

    def _validate_matrix(self, matrix):
        BaseVectorization._validate_matrix(self, matrix)
        self._ensure_real(matrix)

        if not cvxopt_equals(matrix, -matrix.T):
            raise ValueError(
                "The given matrix is not numerically skew-symmetric.")

    def _make_full_to_special(self):
        n = self._shape[0]
        m = n*(n - 1) // 2
        I, J, V = [], [], []
        for i in range(m):
            c = int((1 + 8*i)**0.5 + 1) // 2
            r = i - c*(c - 1) // 2
            I.append(i)
            J.append(c*n + r)
            V.append(2**0.5)
        return cvxopt.spmatrix(V, I, J, (m, n**2), tc="d")

    def _make_special_to_full(self):
        n = self._shape[0]
        I, J, V = [], [], []
        for i in range(n**2):
            c, r = divmod(i, n)
            if c == r:  # Entries on the diagonal are zero.
                continue
            elif c < r:  # Entries below the diagonal are infered.
                I.append(i)
                J.append(r*(r - 1) // 2 + c)
                V.append(-1 / 2**0.5)
            else:
                I.append(i)
                J.append(c*(c - 1) // 2 + r)
                V.append(1 / 2**0.5)
        return cvxopt.spmatrix(V, I, J, (n**2, n*(n - 1) // 2), tc="d")


class HermitianVectorization(BaseVectorization):
    r"""An isometric hermitian matrix vectorization.

    The vectorization is isometric with respect to the Hermitian inner product
    :math:`\langle A, B \rangle = \operatorname{tr}(B^H A)` on the matrices and
    the real dot product on their vectorizations.
    """

    def __init__(self, shape):
        """Initialize a vectorization format for hermitian matrices.

        Uses :class:`SymmetricVectorization` (for the real part) and
        :class:`SkewSymmetricVectorization` (for the imaginary part) internally.
        """
        self._sym = SymmetricVectorization(shape)
        self._skw = SkewSymmetricVectorization(shape)
        super(HermitianVectorization, self).__init__(shape)

    @classmethod
    def _square_input(cls):
        return True

    def _validate_matrix(self, matrix):
        BaseVectorization._validate_matrix(self, matrix)

        if not cvxopt_equals(matrix, matrix.H):
            raise ValueError("The given matrix is not numerically hermitian.")

    def _make_full_to_special(self):
        # No such matrix exists as the full vectorization is complex.
        return None

    def vectorize(self, matrix):
        """Override :meth:`BaseVectorization.vectorize`.

        This is necessary because extracting the real and the imaginary part
        cannot be done with a linear transformation matrix as is done for other
        vectorization formats.
        """
        self._validate_matrix(matrix)
        fullVectorization = matrix[:]
        return cvxopt_vcat([
            self._sym._full2special*fullVectorization.real(),
            self._skw._full2special*fullVectorization.imag()])

    def _make_special_to_full(self):
        A = cvxopt_hcat([
            self._sym._special2full,
            cvxopt.spmatrix([], [], [], self._skw._special2full.size)])
        B = cvxopt_hcat([
            cvxopt.spmatrix([], [], [], self._sym._special2full.size),
            self._skw._special2full])
        return A + 1j*B


class LowerTriangularVectorization(BaseVectorization):
    """An isometric lower triangular matrix vectorization."""

    @classmethod
    def _square_input(cls):
        return True

    def _validate_matrix(self, matrix):
        BaseVectorization._validate_matrix(self, matrix)
        self._ensure_real(matrix)

        n = self._shape[0]
        for i in range(n):
            for j in range(i + 1, n):
                if matrix[i, j]:
                    raise ValueError(
                        "The given matrix is not lower triangular.")

    def _make_full_to_special(self):
        n = self._shape[0]
        m = n*(n + 1) // 2
        I, J, V = [], [], []
        for i in range(m):
            c = int((1 + 8*i)**0.5 - 1) // 2
            r = i - c*(c + 1) // 2
            I.append(i)
            J.append(r*n + c)
            V.append(1)
        return cvxopt.spmatrix(V, I, J, (m, n**2), tc="d")

    def _make_special_to_full(self):
        n = self._shape[0]
        m = n*(n + 1) // 2
        I, J, V = [], [], []
        for j in range(m):
            c = int((1 + 8*j)**0.5 - 1) // 2
            r = j - c*(c + 1) // 2
            I.append(r*n + c)
            J.append(j)
            V.append(1)
        return cvxopt.spmatrix(V, I, J, (n**2, m), tc="d")


class UpperTriangularVectorization(BaseVectorization):
    """An isometric upper triangular matrix vectorization."""

    @classmethod
    def _square_input(cls):
        return True

    def _validate_matrix(self, matrix):
        BaseVectorization._validate_matrix(self, matrix)
        self._ensure_real(matrix)

        n = self._shape[0]
        for i in range(n):
            for j in range(i):
                if matrix[i, j]:
                    raise ValueError(
                        "The given matrix is not upper triangular.")

    def _make_full_to_special(self):
        n = self._shape[0]
        m = n*(n + 1) // 2
        I, J, V = [], [], []
        for i in range(m):
            c = int((1 + 8*i)**0.5 - 1) // 2
            r = i - c*(c + 1) // 2
            I.append(i)
            J.append(c*n + r)
            V.append(1)
        return cvxopt.spmatrix(V, I, J, (m, n**2), tc="d")

    def _make_special_to_full(self):
        n = self._shape[0]
        m = n*(n + 1) // 2
        I, J, V = [], [], []
        for j in range(m):
            c = int((1 + 8*j)**0.5 - 1) // 2
            r = j - c*(c + 1) // 2
            I.append(c*n + r)
            J.append(j)
            V.append(1)
        return cvxopt.spmatrix(V, I, J, (n**2, m), tc="d")


# --------------------------------------
__all__ = api_end(_API_START, globals())



# ------------------------------------------------------------------------------
# Copyright (C) 2019 Maximilian Stahlberg
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Implementation of :class:`EpigraphReformulation`."""

from ..apidoc import api_end, api_start
from ..expressions import AffineExpression, RealVariable
from .reformulation import Reformulation

_API_START = api_start(globals())
# -------------------------------


class EpigraphReformulation(Reformulation):
    """Epigraph reformulation."""

    NEW_OBJECTIVE = AffineExpression.make_type(
        shape=(1, 1), constant=False, nonneg=False)

    @classmethod
    def supports(cls, footprint):
        """Implement :meth:`~.reformulation.Reformulation.supports`."""
        return footprint.objective.clstype is not AffineExpression

    @classmethod
    def predict(cls, footprint):
        """Implement :meth:`~.reformulation.Reformulation.predict`."""
        if footprint.direction == "max":
            constraint = footprint.objective >= cls.NEW_OBJECTIVE
        else:
            constraint = footprint.objective <= cls.NEW_OBJECTIVE

        return footprint.updated((
            ("obj", footprint.NONE),
            ("obj", cls.NEW_OBJECTIVE, None),
            ("var", RealVariable.make_var_type(dim=1, bnd=0), 1),
            ("con", constraint, 1)))

    def forward(self):
        """Implement :meth:`~.reformulation.Reformulation.forward`."""
        self.output = self.input.clone(copyOptions=False)

        direction, objective = self.output.objective

        self.t = self.output.add_variable("__t")
        self.C = self.output.add_constraint(
            objective <= self.t if direction == "min" else objective >= self.t)
        self.output.set_objective(direction, self.t)

    def update(self):
        """Implement :meth:`~.reformulation.Reformulation.update`."""
        if self._objective_has_changed():
            # Remove the old auxilary constraint.
            self.output.remove_constraint(self.C.id)

            # Add a new one, using the existing variable.
            newDir, newObj = self.input.objective
            self.C = self.output.add_constraint(
                newObj <= self.t if newDir == "min" else newObj >= self.t)

        self._pass_updated_vars()
        self._pass_updated_cons()
        self._pass_updated_options()

    def backward(self, solution):
        """Implement :meth:`~.reformulation.Reformulation.backward`."""
        if self.t in solution.primals:
            solution.primals.pop(self.t)

        if self.C in solution.duals:
            solution.duals.pop(self.C)

        return solution


# --------------------------------------
__all__ = api_end(_API_START, globals())



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2019 Maximilian Stahlberg
# Based on the original picos.expressions module by Guillaume Sagnol.
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Mathematical expression types."""

from ..apidoc import api_end, api_start

_API_START = api_start(globals())
# -------------------------------


from .algebra import *  # noqa
from .exp_affine import (AffineExpression, ComplexAffineExpression,  # noqa
                         Constant)
from .exp_detrootn import DetRootN  # noqa
from .exp_entropy import Entropy, NegativeEntropy  # noqa
from .exp_geomean import GeometricMean  # noqa
from .exp_logarithm import Logarithm  # noqa
from .exp_logsumexp import LogSumExp  # noqa
from .exp_norm import Norm, SpectralNorm, NuclearNorm  # noqa
from .exp_powtrace import PowerTrace  # noqa
from .exp_quadratic import QuadraticExpression  # noqa
from .exp_sumexp import SumExponentials  # noqa
from .exp_sumxtr import SumExtremes  # noqa
from .expression import Expression, ExpressionType, NotValued  # noqa
from .variables import (BaseVariable, BinaryVariable, ComplexVariable,  # noqa
                        HermitianVariable, IntegerVariable,
                        LowerTriangularVariable, RealVariable,
                        SkewSymmetricVariable, SymmetricVariable,
                        UpperTriangularVariable, CONTINUOUS_VARTYPES)
from .set import Set, SetType  # noqa
from .set_ball import Ball  # noqa
from .set_expcone import ExponentialCone  # noqa
from .set_simplex import Simplex  # noqa
from .set_soc import SecondOrderCone  # noqa
from .set_rsoc import RotatedSecondOrderCone  # noqa


# --------------------------------------
__all__ = api_end(_API_START, globals())



# ------------------------------------------------------------------------------
# Copyright (C) 2019 Maximilian Stahlberg
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Implements a helper reformulation to apply temporary options."""

from ..apidoc import api_end, api_start
from .reformulation import Reformulation

_API_START = api_start(globals())
# -------------------------------


class ExtraOptions(Reformulation):
    """Helper reformulation to apply temporary options.

    This reformulation is different from all others in a number of ways:

    - It doesn't change the footprint of any problem.
    - It is automatically the first reformulation in any strategy.
    - It is the only reformulation whose :meth:`execute` accepts a keyword
      argument sequence of additional options to use.
    - It is the only reformulation that can be skipped entirely (by setting
      ``self.output = self.input``).

    The job of this reformulation is to apply temporary options passed to
    :meth:`Problem.solve <picos.Problem.solve>` so that subsequent
    reformulations can find their options stored in their input problem.
    """

    @classmethod
    def supports(cls, footprint):
        """Implement :meth:`~.reformulation.Reformulation.supports`."""
        return True

    @classmethod
    def predict(cls, footprint):
        """Implement :meth:`~.reformulation.Reformulation.predict`."""
        # The additional options are already part of the footprint!
        # (They are applied before the first prediction takes place.)
        return footprint

    def execute(self, **extra_options):
        """Override :meth:`~.reformulation.Reformulation.execute`.

        Adds the ``extra_options`` argument and attempts to perform as little
        reformulation work as possible.
        """
        newOptions = self.input.options.self_or_updated(**extra_options)
        verbose = newOptions.verbosity > 0

        if newOptions == self.input.options:
            # No options need to be applied; shortcut this reformulation so that
            # no forwarding or updating work needs to be done.
            if verbose:
                print("Skipping {}.".format(self.__class__.__name__))

            # Shortcut the whole problem.
            self.output = self.input

            self._reset_knowns()
        elif self.output and self.output is not self.input:
            # We made a problem clone before and we still can't shortcut, so
            # perform a regular update and apply the new options to the copy.
            if verbose:
                print("Updating {}.".format(self.__class__.__name__))

            # Update the options.
            self.output.options = newOptions

            self._pass_updated_objective()
            self._pass_updated_vars()
            self._pass_updated_cons()
        else:
            # We need to apply temporary options and this is either the first
            # execution or the previous execution was a shortcut. Perform a
            # regular forward and apply the new options to the problem clone.
            if verbose:
                print("Applying {}.".format(self.__class__.__name__))

            # Copy the problem and update the options.
            self.output = self.input.clone(copyOptions=False)
            self.output.options = newOptions

            self._set_knowns()

        # Make sure there is a successor to obtain a solution from.
        assert self.successor, \
            "The reformulation being executed has no successor."

        # Advance one step and return any solution as-is.
        return self.successor.execute()

    def forward(self):
        """Dummy-implement :meth:`~.reformulation.Reformulation.forward`."""
        pass

    def update(self):
        """Dummy-implement :meth:`~.reformulation.Reformulation.update`."""
        pass

    def backward(self, solution):
        """Dummy-implement :meth:`~.reformulation.Reformulation.backward`."""
        pass


# --------------------------------------
__all__ = api_end(_API_START, globals())



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2019 Maximilian Stahlberg
# Based on the original picos.expressions module by Guillaume Sagnol.
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Backend for expression type implementations."""

import functools
import operator
import sys

import cvxopt

from .. import glyphs
from ..apidoc import api_end, api_start
from ..compat import ABC, abstractmethod
from ..constraints import ConstraintType
from ..containers import DetailedType
from ..legacy import deprecated

from .data import convert_operands

_API_START = api_start(globals())
# -------------------------------


class NotValued(RuntimeError):
    """The operation cannot be performed due to a variable without a value.

    Note that the :attr:`~picos.expressions.Expression.value` and
    :attr:`~picos.expressions.Expression.value_as_matrix` attributes do not
    raise this exception, but return :obj:`None`.
    """

    pass


class PredictedFailure(TypeError):
    """Denotes that comparing two expressions will not form a constraint."""

    pass


class ExpressionType(DetailedType):
    """The detailed type of an expression for predicting constraint outcomes.

    This is suffcient to predict the detailed type of any constraint that can be
    created by comparing with another expression.

    The prediction is done by using the relevant operator on instances of this
    class, as opposed to the expressions themselves. Note that operators used to
    create new expressions as opposed to constraints are not handled.
    """

    @staticmethod
    def _relation_str(relation):
        if relation is operator.__eq__:
            return "=="
        elif relation is operator.__le__:
            return "<="
        elif relation is operator.__ge__:
            return ">="
        elif relation is operator.__lshift__:
            return "<<"
        elif relation is operator.__rshift__:
            return ">>"
        else:
            return "??"

    @staticmethod
    def _swap_relation(relation):
        if relation is operator.__eq__:
            return operator.__eq__
        elif relation is operator.__le__:
            return operator.__ge__
        elif relation is operator.__ge__:
            return operator.__le__
        else:
            return None

    def _compare(self, relation, other):
        if not isinstance(other, ExpressionType):
            return NotImplemented  # Results in a regular Python TypeError.

        # Perform the forward prediction.
        result = self.clstype._predict(self.subtype, relation, other)

        # Fall back to the backward prediction.
        if result is NotImplemented:
            reverse = self._swap_relation(relation)
            result  = other.clstype._predict(other.subtype, reverse, self)

        # If both fail, the prediction is "not possible".
        if result is NotImplemented:
            raise PredictedFailure(
                "The statement {} {} {} is predicted to error."
                .format(self, self._relation_str(relation), other))
        else:
            assert isinstance(result, ConstraintType)
            return result

    def __eq__(self, other):
        return self._compare(operator.__eq__, other)

    def __le__(self, other):
        return self._compare(operator.__le__, other)

    __lt__ = __le__

    def __ge__(self, other):
        return self._compare(operator.__ge__, other)

    __gt__ = __ge__

    def __lshift__(self, other):
        return self._compare(operator.__lshift__, other)

    def __rshift__(self, other):
        return self._compare(operator.__rshift__, other)


class Expression(ABC):
    """Abstract base class for mathematical expressions, including variables.

    For variables, this is the secondary base class, with
    :class:`~.variables.BaseVariable` being the primary one.
    """

    def __init__(self, typeStr, symbStr):
        """Perform basic initialization for :class:`Expression` instances.

        :param str typeStr: Short string denoting the expression type.
        :param str symbStr: Algebraic string description of the expression.
        """
        self._typeStr = typeStr
        """A string describing the expression type."""

        self._symbStr = symbStr
        """A symbolic string representation of the expression. It is always used
        by __descr__, and it is equivalent to the value returned by __str__ when
        the expression is not fully valued."""

    @property
    def string(self):
        """Symbolic string representation of the expression.

        Use this over Python's :class:`str` if you want to output the symbolic
        representation even when the expression is valued.
        """
        return self._symbStr

    # --------------------------------------------------------------------------
    # Abstract and default-implementation methods.
    # --------------------------------------------------------------------------

    def _get_refined(self):
        """See :attr:`refined`."""
        return self

    def _get_clstype(self):
        """Return the Python class part of the expression's detailed type."""
        return self.__class__

    @property
    @abstractmethod
    def Subtype(self):
        """The class of which :attr:`subtype` returns an instance.

        Instances must be hashable. By convention a
        :func:`namedtuple <collections.namedtuple>` class.

        .. warning::
            This should be declared in the class body as e.g.
            `Subtype = namedtuple(…)` and not as a property so that it's static.
        """
        pass

    @abstractmethod
    def _get_subtype(self):
        """See :attr:`subtype`."""
        pass

    @classmethod
    @abstractmethod
    def _predict(cls, subtype, relation, other):
        """Predict the constraint outcome of a comparison.

        :param object subtype: An object returned by the :meth:`_get_subtype`
            instance method of :class:`cls`.
        :param method-wrapper relation: A function from the :mod:`operator`
            namespace, such as :func:`operator.__le__`. See
            :class:`ExpressionType` for what operators are defined.
        :param ExpressionType other: The detailed type of another expression.
        :returns: Either the :obj:`NotImplemented` token or a
            :class:`ConstraintType` object such that an instance of :class:`cls`
            with the given subtype, when compared with another expression with
            the given expression type, returns a constraint with that constraint
            type.
        """
        pass

    @abstractmethod
    def _get_value(self):
        """Return the numeric value of the expression as a CVXOPT matrix.

        :raises NotValued: When the value is not fully defined.

        Method implementations need to return an independent copy of the value
        that the user is allowed to change without affecting the expression.
        """
        pass

    def _set_value(self, value):
        raise NotImplementedError("Setting the value on an instance of {} is "
            "not supported, but you can value any variables involved instead."
            .format(type(self).__name__))

    def _get_shape(self):
        """Return the algebraic shape of the expression."""
        return (1, 1)

    @abstractmethod
    def _get_variables(self):
        """Return the set of variables that are involved in the expression."""
        pass

    @abstractmethod
    def _is_convex(self):
        """Whether the expression is convex."""
        pass

    @abstractmethod
    def _is_concave(self):
        """Whether the expression is concave."""
        pass

    @abstractmethod
    def _replace_variables(self, var_map):
        """Return a copy of the expression concerning different variables.

        This is the fast internal-use counterpart to :meth:`replace_variables`.

        :param dict var_map:
            Maps old to old or new variables. The map must be complete and sane,
            that is every variable involved in the expression must occur in the
            map and be mapped either to itself or another variable of same shape
            and vectorization format.
        """
        pass

    # --------------------------------------------------------------------------
    # An interface to the abstract and default-implementation methods above.
    # --------------------------------------------------------------------------

    def replace_variables(self, new_variables):
        """Return a copy of the expression concerning different variables.

        New variables must have the same shape and vectorization format as the
        variables that they replace. This means in particular that
        :class:`~.variables.RealVariable`, :class:`~.variables.IntegerVariable`
        and :class:`~.variables.BinaryVariable` of same shape are
        interchangeable.

        If the variables to be replaced do not appear in the expression, then
        the expression is not copied but returned as is.

        :param new_variables:
            Either a map from variables or variable names to new variables or an
            iterable of new variables to replace existing variables of same name
            with.
        :type new_variables:
            tuple or list or dict
        """
        old_variables = {var.name: var for var in self.variables}

        # Check unique naming of existing variables.
        if len(old_variables) != len(self.variables):
            raise RuntimeError("Cannot perform variable replacement on {} as "
                "its variables are not uniquely named.".format(self.string))

        # Obtain an existing variable to new variable mapping.
        if isinstance(new_variables, dict):
            # Replace keys that are variables with their respective names.
            new_new_variables = {
                (old if isinstance(old, str) else old.name): new_var
                for old, new_var in new_variables.items()}
            if len(new_new_variables) != len(new_variables):
                raise ValueError("New variables must have unique names.")
            new_variables = new_new_variables

            replacement = {old_variables[common_name]: new_var
                for common_name, new_var in new_variables.items()
                if common_name in old_variables}
        else:
            replacement = {old_variables[new_var.name]: new_var
                for new_var in new_variables if new_var.name in old_variables}

        if not replacement:
            return self

        # Validate replacement requirements.
        for old_var, new_var in replacement.items():
            if old_var.shape != new_var.shape:
                raise TypeError(
                    "Cannot replace {} with {} in {}: Differing shape."
                    .format(old_var.name, new_var.name, self.string))
            elif type(old_var._vec) != type(new_var._vec):  # noqa: E721
                raise TypeError(
                    "Cannot replace {} with {} in {}: Differing vectorization."
                    .format(old_var.name, new_var.name, self.string))

        var_map = {var: var for var in self.variables}
        var_map.update(replacement)

        return self._replace_variables(var_map)

    @property
    def refined(self):
        """A refined version of the expression.

        The refined expression can be an instance of a different
        :class:`Expression` subclass than the original expression, if that type
        is better suited for the mathematical object in question.

        The refined expression is automatically used instead of the original one
        whenever a constraint is created, and in some other places.

        The idea behind refined expressions is that operations that produce new
        expressions can be executed quickly without checking for exceptionnel
        cases. For instance, the sum of two
        :class:`~.exp_affine.ComplexAffineExpression` instances could have the
        complex part eliminated so that storing the result as an
        :class:`~.exp_affine.AffineExpression` would be prefered, but checking
        for this case on every addition would be too slow. Refinement is used
        sparingly to detect such cases at times where it makes the most sense.
        """
        fine = self._get_refined()

        if fine is not self:
            # Recursively refine until the expression doesn't change further.
            return fine.refined
        else:
            return fine

    @property
    def subtype(self):
        """The subtype part of the expression's detailed type.

        Returns a hashable object that, together with the Python class part of
        the expression's type, is sufficient to predict the constraint outcome
        (constraint class and subtype) of any comparison operation with any
        other expression.

        By convention the object returned is a
        :func:`namedtuple <collections.namedtuple>` instance.
        """
        return self._get_subtype()

    @property
    def type(self):
        """The expression's detailed type for constraint prediction.

        The returned value is suffcient to predict the detailed type of any
        constraint that can be created by comparing with another expression.

        Since constraints are created from
        :attr:`~picos.expressions.Expression.refined` expressions only, the
        Python class part of the detailed type may differ from the type of the
        expression whose :attr:`type` is queried.
        """
        refined = self.refined
        return ExpressionType(refined._get_clstype(), refined._get_subtype())

    @classmethod
    def make_type(cls, *args, **kwargs):
        """Create a detailed expression type from subtype parameters."""
        return ExpressionType(cls, cls.Subtype(*args, **kwargs))

    def _wrap_get_value(self, asMatrix):
        """Enhance the implementation of :attr:`_get_value`.

        This turns :exc:`NotValued` exceptions into a return value of
        :obj:`None` and checks the type of any value returned.

        :param bool asMatrix: Whether scalar values are returned as matrices.
        """
        try:
            value = self._get_value()
        except NotValued:
            return None

        assert isinstance(value, (cvxopt.matrix, cvxopt.spmatrix)), \
            "Expression._get_value implementations must return a CVXOPT matrix."

        if value.size == (1, 1) and not asMatrix:
            return value[0]
        else:
            return value

    value = property(
        lambda self: self._wrap_get_value(asMatrix=False),
        lambda self, x: self._set_value(x),
        lambda self: self._set_value(None),
        r"""Value of the expression.

        It is defined (not :obj:`None`) if the expression is constant or if all
        variables involved in the expression are valued. Variables can be valued
        directly by writing to their :attr:`value` attribute, or they can be
        valued by PICOS when an optimization solution is found.

        Some expressions can also be valued directly if PICOS can find a minimal
        norm variable assignment that makes the expression have the desired
        value. In particular, this works with affine expressions whose linear
        part has an under- or well-determined coefficient matrix.

        :Example:

        >>> from picos import RealVariable
        >>> x = RealVariable("x", (1,3))
        >>> y = RealVariable("y", (1,3))
        >>> e = x - 2*y + 3
        >>> print("e:", e)
        e: x - 2·y + [3]
        >>> e.value = [4, 5, 6]
        >>> print("e: ", e, "\nx: ", x, "\ny: ", y, sep = "")
        e: [ 4.00e+00  5.00e+00  6.00e+00]
        x: [ 2.00e-01  4.00e-01  6.00e-01]
        y: [-4.00e-01 -8.00e-01 -1.20e+00]

        Unlike :attr:`value_as_matrix`, scalars are returned as scalar types.
        """)

    value_as_matrix = property(
        lambda self: self._wrap_get_value(asMatrix=True),
        lambda self, x: self._set_value(x),
        lambda self: self._set_value(None),
        """Value of the expression, as a CVXOPT matrix type.

        Refer to :attr:`value` for when it is defined (not :obj:`None`).

        Unlike :attr:`value`, scalars are returned in the form of 1x1 matrices.
        """)

    @property
    def valued(self):
        """Whether the expression is valued.

        .. note::

            Querying this attribute is *not* faster than immediately querying
            :attr:`value` and checking whether it is :obj:`None`. Use it only if
            you do not need to know the value, but only whether it is available.

        :Example:

        >>> from picos import RealVariable
        >>> x = RealVariable("x", 3)
        >>> x.valued
        False
        >>> x.value
        >>> print((x|1))
        ∑(x)
        >>> x.value = [1, 2, 3]
        >>> (x|1).valued
        True
        >>> print((x|1))
        6.0
        """
        try:
            self._get_value()
        except NotValued:
            return False
        else:
            return True

    @valued.setter
    def valued(self, x):
        if x is False:
            self._set_value(None)
        else:
            raise ValueError("You may only assign 'False' to the 'valued' "
                "attribute, which is the same as setting 'value' to 'None'.")

    shape = property(
        lambda self: self._get_shape(),
        doc=_get_shape.__doc__)

    size = property(
        lambda self: self._get_shape(),
        doc="""The same as :attr:`shape`.""")

    variables = property(
        lambda self: self._get_variables(),
        doc=_get_variables.__doc__)

    convex = property(
        lambda self: self._is_convex(),
        doc=_is_convex.__doc__)

    concave = property(
        lambda self: self._is_concave(),
        doc=_is_concave.__doc__)

    # --------------------------------------------------------------------------
    # Python special method implementations and their helpers.
    # --------------------------------------------------------------------------

    def __len__(self):
        return self.shape[0] * self.shape[1]

    def __le__(self, other):
        # Try to refine self and see if the operation is then supported.
        # This allows e.g. a <= 0 if a is a real-valued complex expression.
        refined = self.refined
        if type(refined) != type(self):
            return refined.__le__(other)

        # HACK: Mimic Python 3's exception when neither a.__le__(b) nor
        #       b.__ge__(a) are implemented.
        if sys.version_info[0] == 2:
            setattr(self, "LT_HACK_TOKEN", None)
            try:
                if hasattr(other, "__ge__") \
                and not hasattr(other, "GT_HACK_TOKEN"):
                    retval = other.__ge__(self)
                    if retval is not NotImplemented:
                        return retval

                raise TypeError(
                    "Constraints of the form '{} <= {}' are not supported."
                    .format(type(self).__name__, type(other).__name__))
            finally:
                delattr(self, "LT_HACK_TOKEN")
        else:
            return NotImplemented

    def __ge__(self, other):
        # Try to refine self and see if the operation is then supported.
        # This allows e.g. a >= 0 if a is a real-valued complex expression.
        refined = self.refined
        if type(refined) != type(self):
            return refined.__ge__(other)

        # HACK: Mimic Python 3's exception when neither a.__le__(b) nor
        #       b.__ge__(a) are implemented.
        if sys.version_info[0] == 2:
            setattr(self, "GT_HACK_TOKEN", None)
            try:
                if hasattr(other, "__le__") \
                and not hasattr(other, "LT_HACK_TOKEN"):
                    retval = other.__le__(self)
                    if retval is not NotImplemented:
                        return retval

                raise TypeError(
                    "Constraints of the form '{} >= {}' are not supported."
                    .format(type(self).__name__, type(other).__name__))
            finally:
                delattr(self, "GT_HACK_TOKEN")
        else:
            return NotImplemented

    def __invert__(self):
        """Convert between a valued expression and its value.

        The value is returned as a constant affine expression whose conversion
        returns the original expression.
        """
        if hasattr(self, "_origin"):
            return self._origin
        elif self.constant:
            return self

        from .exp_affine import Constant

        A = Constant(glyphs.value(self.string), self._get_value(), self.shape)
        A._origin = self
        return A

    def __contains__(self, variable):
        """Whether the expression concerns the given variable."""
        return variable in self.variables

    def __eq__(self, exp):
        raise NotImplementedError("PICOS supports equality comparison only "
            "between affine expressions, as otherwise the problem would "
            "become non-convex. Choose either <= or >= if possible.")

    def __repr__(self):
        return str(glyphs.repr2(self._typeStr, self._symbStr))

    def __str__(self):
        """Return a dynamic string description of the expression.

        The description is based on whether the expression is valued. If it is
        valued, then a string representation of the value is returned.
        Otherwise, the symbolic description of the expression is returned.
        """
        if self.valued:
            return str(self.value).strip()
        else:
            return str(self._symbStr)

    def __format__(self, format_spec):
        if self.valued:
            return self.value.__format__(format_spec)
        else:
            return self._symbStr.__format__(format_spec)

    def __index__(self):
        if len(self) != 1:
            raise TypeError("Cannot use multidimensional expression {} as an "
                "index.".format(self.string))

        if not self.valued:
            raise NotValued("Cannot use unvalued expression {} as an index."
                .format(self.string))

        value = self.value

        try:
            fltValue = float(value)
            intValue = int(fltValue)

            if intValue != fltValue:
                raise ValueError
        except (TypeError, ValueError):
            raise RuntimeError("Cannot use {} as an index as its value of {} is"
                " not integral.".format(self.string, value))

        return intValue

    def _casting_helper(self, theType):
        assert theType in (int, float, complex)

        if len(self) != 1:
            raise TypeError(
                "Cannot cast multidimensional expression {} as {}."
                .format(self.string, theType.__name__))

        if not self.valued:
            raise NotValued("Cannot cast unvalued expression {} as {}."
                .format(self.string, theType.__name__))

        value = self.value_as_matrix

        return theType(value[0])

    def __int__(self):
        return self._casting_helper(int)

    def __long__(self):
        # Python 2 only.
        return self._casting_helper(int).__long__()

    def __float__(self):
        return self._casting_helper(float)

    def __complex__(self):
        return self._casting_helper(complex)

    def __round__(self, ndigits=None):
        return round(float(self), ndigits)

    # Since we define __eq__, __hash__ is not inherited. Do this manually.
    __hash__ = object.__hash__

    # HACK: This prevents NumPy operators from iterating over PICOS expressions.
    __array_priority__ = float("inf")

    # --------------------------------------------------------------------------
    # Fallback algebraic operations: Try again with converted RHS, refined LHS.
    # --------------------------------------------------------------------------

    @convert_operands(sameShape=True)
    def __add__(self, other):
        if type(self.refined) != type(self):
            return self.refined.__add__(other)
        else:
            return NotImplemented

    @convert_operands(sameShape=True)
    def __radd__(self, other):
        if type(self.refined) != type(self):
            return self.refined.__radd__(other)
        else:
            return NotImplemented

    @convert_operands(sameShape=True)
    def __sub__(self, other):
        if type(self.refined) != type(self):
            return self.refined.__sub__(other)
        else:
            return NotImplemented

    @convert_operands(sameShape=True)
    def __rsub__(self, other):
        if type(self.refined) != type(self):
            return self.refined.__rsub__(other)
        else:
            return NotImplemented

    @convert_operands(sameShape=True)
    def __or__(self, other):
        if type(self.refined) != type(self):
            return self.refined.__or__(other)
        else:
            return NotImplemented

    @convert_operands(sameShape=True)
    def __ror__(self, other):
        if type(self.refined) != type(self):
            return self.refined.__ror__(other)
        else:
            return NotImplemented

    @convert_operands(rMatMul=True)
    def __mul__(self, other):
        if type(self.refined) != type(self):
            return self.refined.__mul__(other)
        else:
            return NotImplemented

    @convert_operands(lMatMul=True)
    def __rmul__(self, other):
        if type(self.refined) != type(self):
            return self.refined.__rmul__(other)
        else:
            return NotImplemented

    @convert_operands(sameShape=True)
    def __xor__(self, other):
        if type(self.refined) != type(self):
            return self.refined.__xor__(other)
        else:
            return NotImplemented

    @convert_operands(sameShape=True)
    def __rxor__(self, other):
        if type(self.refined) != type(self):
            return self.refined.__rxor__(other)
        else:
            return NotImplemented

    @convert_operands()
    def __matmul__(self, other):
        if type(self.refined) != type(self):
            return self.refined.__matmul__(other)
        else:
            return NotImplemented

    @convert_operands()
    def __rmatmul__(self, other):
        if type(self.refined) != type(self):
            return self.refined.__rmatmul__(other)
        else:
            return NotImplemented

    @convert_operands(scalarRHS=True)
    def __truediv__(self, other):
        if type(self.refined) != type(self):
            return self.refined.__truediv__(other)
        else:
            return NotImplemented

    @convert_operands(scalarLHS=True)
    def __rtruediv__(self, other):
        if type(self.refined) != type(self):
            return self.refined.__rtruediv__(other)
        else:
            return NotImplemented

    @convert_operands(scalarRHS=True)
    def __div__(self, other):
        if type(self.refined) != type(self):
            return self.refined.__div__(other)
        else:
            return NotImplemented

    @convert_operands(scalarLHS=True)
    def __rdiv__(self, other):
        if type(self.refined) != type(self):
            return self.refined.__rdiv__(other)
        else:
            return NotImplemented

    @convert_operands(scalarRHS=True)
    def __pow__(self, other):
        if type(self.refined) != type(self):
            return self.refined.__pow__(other)
        else:
            return NotImplemented

    @convert_operands(scalarLHS=True)
    def __rpow__(self, other):
        if type(self.refined) != type(self):
            return self.refined.__rpow__(other)
        else:
            return NotImplemented

    @convert_operands(horiCat=True)
    def __and__(self, other):
        if type(self.refined) != type(self):
            return self.refined.__and__(other)
        else:
            return NotImplemented

    @convert_operands(horiCat=True)
    def __rand__(self, other):
        if type(self.refined) != type(self):
            return self.refined.__rand__(other)
        else:
            return NotImplemented

    @convert_operands(vertCat=True)
    def __floordiv__(self, other):
        if type(self.refined) != type(self):
            return self.refined.__floordiv__(other)
        else:
            return NotImplemented

    @convert_operands(vertCat=True)
    def __rfloordiv__(self, other):
        if type(self.refined) != type(self):
            return self.refined.__rfloordiv__(other)
        else:
            return NotImplemented

    def __neg__(self):
        if type(self.refined) != type(self):
            return self.refined.__neg__()
        else:
            return NotImplemented

    def __abs__(self):
        if type(self.refined) != type(self):
            return self.refined.__abs__()
        else:
            return NotImplemented

    # --------------------------------------------------------------------------
    # Backwards compatibility methods.
    # --------------------------------------------------------------------------

    @deprecated("2.0", useInstead="valued")
    def is_valued(self):
        """Whether the expression is valued."""
        return self.valued

    @deprecated("2.0", useInstead="value")
    def set_value(self, value):
        """Set the value of an expression, usually a variable."""
        self.value = value

    @deprecated("2.0", "PICOS treats all inequalities as non-strict. Using the "
        "strict inequality comparison operators may lead to unexpected results "
        "when dealing with integer problems.")
    def __lt__(self, exp):
        return self.__le__(exp)

    @deprecated("2.0", "PICOS treats all inequalities as non-strict. Using the "
        "strict inequality comparison operators may lead to unexpected results "
        "when dealing with integer problems.")
    def __gt__(self, exp):
        return self.__ge__(exp)


def validate_prediction(operator):
    """Validate that the constraint outcome matches the predicted outcome."""
    @functools.wraps(operator)
    def wrapper(lhs, rhs, *args, **kwargs):
        from .set import Set

        def what():
            return "({}).{}({})".format(
                lhs._symbStr, operator.__name__, rhs._symbStr)

        assert isinstance(lhs, (Expression, Set)) \
            and isinstance(rhs, (Expression, Set)), \
            "validate_prediction must occur below convert_operands."

        lhs_type = lhs.type

        assert hasattr(lhs_type, operator.__name__), "Constraint outcome " \
            "prediction is not implemented for the operation {}." \
            .format(operator.__name__)

        try:
            predictedType = getattr(lhs_type, operator.__name__)(rhs.type)
        except NotImplementedError:
            predictedType = None  # No prediction was made.
        except PredictedFailure:
            predictedType = NotImplemented  # Prediction is "not possible".
        except Exception as error:
            raise AssertionError("Cannot validate the prediciton for {}: {}"
                .format(what(), error))

        try:
            outcome = operator(lhs, rhs, *args, **kwargs)
        except Exception as error:
            # Case where the prediction is positive and the outcome is negative.
            if predictedType not in (None, NotImplemented):
                raise AssertionError(
                    "Outcome for {} was predicted {} but the operation raised "
                    "an error: {}".format(what(), predictedType, error))
            else:
                raise

        # Case where the prediction is negative and the outcome is positive.
        if predictedType is NotImplemented and outcome is not NotImplemented:
            raise AssertionError(
                "The operation {} was predicted to fail but it produced "
                "an output of {}.".format(what(), outcome.type))

        # Case where no prediction was made.
        if not predictedType:
            return outcome

        # Case where the outcome is try-to-reverse-the-operation.
        if outcome is NotImplemented:
            return outcome

        # Case where the prediction and the outcome are positive but differ.
        outcomeType = outcome.type
        if not predictedType.equals(outcomeType):
            raise AssertionError("Outcome for {} was predicted {} but is {}."
                .format(what(), predictedType, outcomeType))

        return outcome
    return wrapper


def refine_operands(stop_at_affine=False):
    """Cast :meth:`~Expression.refined` on both operands.

    If the left hand side operand (i.e. ``self``) is refined to an instance of a
    different type, then, instead of the decorated method, the method with the
    same name on the refined type is invoked with the (refined) right hand side
    operand as its argument.

    This decorator is supposed to be used on all constraint creating binary
    operator methods so that degenerated instances (e.g. a complex affine
    expression with an imaginary part of zero) can occur but are not used in
    constraints. This speeds up many computations involving expressions as these
    degenerate cases do not need to be detected. Note that
    :attr:`Expression.type` also refers to the refined version of an expression.

    :param bool stop_at_affine: Do not refine any affine expressions, in
        particular do not refine complex affine expressions to real ones.
    """
    def decorator(operator):
        @functools.wraps(operator)
        def wrapper(lhs, rhs, *args, **kwargs):
            from .set import Set
            from .exp_affine import ComplexAffineExpression

            assert isinstance(lhs, (Expression, Set)) \
                and isinstance(rhs, (Expression, Set)), \
                "refine_operands must occur below convert_operands."

            if stop_at_affine and isinstance(lhs, ComplexAffineExpression):
                lhs_refined = lhs
            else:
                lhs_refined = lhs.refined

            if type(lhs_refined) is not type(lhs):
                assert hasattr(lhs_refined, operator.__name__), \
                    "refine_operand transformed 'self' to another type that " \
                    "does not define an operator with the same name as the " \
                    "decorated one."

                refined_operation = getattr(lhs_refined, operator.__name__)

                return refined_operation(rhs, *args, **kwargs)

            if stop_at_affine and isinstance(rhs, ComplexAffineExpression):
                rhs_refined = rhs
            else:
                rhs_refined = rhs.refined

            return operator(lhs_refined, rhs_refined, *args, **kwargs)
        return wrapper
    return decorator


# --------------------------------------
__all__ = api_end(_API_START, globals())



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2019 Maximilian Stahlberg
# Based on parts of the tools module by Guillaume Sagnol.
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Implements functions that create or modify algebraic expressions."""

import functools

from .. import glyphs
from ..apidoc import api_end, api_start
from ..compat import builtins
from ..formatting import parameterized_string
from ..legacy import deprecated, throw_deprecation_warning
from .data import convert_and_refine_arguments
from .exp_affine import ComplexAffineExpression, Constant
from .exp_detrootn import DetRootN
from .exp_entropy import NegativeEntropy
from .exp_geomean import GeometricMean
from .exp_logsumexp import LogSumExp
from .exp_norm import Norm
from .exp_powtrace import PowerTrace
from .exp_sumexp import SumExponentials
from .exp_sumxtr import SumExtremes
from .expression import Expression
from .set_ball import Ball
from .set_expcone import ExponentialCone
from .set_rsoc import RotatedSecondOrderCone
from .set_simplex import Simplex
from .set_soc import SecondOrderCone

_API_START = api_start(globals())
# -------------------------------


# ------------------------------------------------------------------------------
# Algebraic functions with a logic of their own.
# ------------------------------------------------------------------------------


def sum(lst, it=None, indices=None):
    """Sum PICOS expressions and give the result a meaningful description.

    This is a replacement for Python's :func:`sum` that produces sensible string
    representations when summing PICOS expressions.

    :param lst: A list of
        :class:`~picos.expressions.ComplexAffineExpression`, or a
        single affine expression whose elements shall be summed.
    :type lst: list or tuple or
        ~picos.expressions.ComplexAffineExpression

    :param it: DEPRECATED
    :param indices: DEPRECATED

    :Example:

    >>> import builtins
    >>> import picos
    >>> x = picos.RealVariable("x", 5)
    >>> e = [x[i]*x[i+1] for i in range(len(x) - 1)]
    >>> builtins.sum(e)
    <Quadratic Expression: x[0]·x[1] + x[1]·x[2] + x[2]·x[3] + x[3]·x[4]>
    >>> picos.sum(e)
    <Quadratic Expression: ∑(x[i]·x[i+1] : i ∈ [0…3])>
    >>> picos.sum(x)  # The same as (x|1).
    <1×1 Real Linear Expression: ∑(x)>
    """
    if it is not None or indices is not None:
        # Deprecated as of 2.0.
        throw_deprecation_warning("Arguments 'it' and 'indices' to picos.sum "
            "are deprecated and ignored.")

    if isinstance(lst, Expression):
        if isinstance(lst, ComplexAffineExpression):
            return (lst | 1.0)
        else:
            raise TypeError(
                "PICOS doesn't know how to sum over a single {}."
                .format(type(lst).__name__))

    if any(not isinstance(expression, Expression) for expression in lst):
        return builtins.sum(lst)

    if len(lst) == 0:
        return Constant(0)
    elif len(lst) == 1:
        return lst[0]
    elif len(lst) == 2:
        return lst[0] + lst[1]

    theSum = lst[0]
    for expression in lst[1:]:
        theSum += expression

    try:
        template, data = parameterized_string([exp.string for exp in lst])
    except ValueError:
        theSum._symbStr = glyphs.sum(
            glyphs.fromto(lst[0].string + ", ", ", " + lst[-1].string))
    else:
        theSum._symbStr = glyphs.sum(glyphs.sep(template, data))

    return theSum


# ------------------------------------------------------------------------------
# Functions that call expression methods.
# ------------------------------------------------------------------------------


def _error_on_none(func):
    """Raise a :exc:`TypeError` if the function returns :obj:`None`."""
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        result = func(*args, **kwargs)

        if result is None:
            raise TypeError("PICOS does not have a representation for {}({})."
                .format(func.__qualname__ if hasattr(func, "__qualname__") else
                func.__name__, ", ".join([type(x).__name__ for x in args] +
                ["{}={}".format(k, type(x).__name__) for k, x in kwargs.items()]
                )))

        return result
    return wrapper


@_error_on_none
@convert_and_refine_arguments("x")
def exp(x):
    """Denote the exponential."""
    if hasattr(x, "exp"):
        return x.exp


@_error_on_none
@convert_and_refine_arguments("x")
def log(x):
    """Denote the natural logarithm."""
    if hasattr(x, "log"):
        return x.log


@_error_on_none
@convert_and_refine_arguments("x", "y")
def kron(x, y):
    """Denote the kronecker product.

    Note that with Python 3, you can just write ``x @ y``.
    """
    if hasattr(x, "kron"):
        return x.kron(y)


@_error_on_none
@convert_and_refine_arguments("x")
def diag(x, n=1):
    r"""Form a diagonal matrix from the column-major vectorization of :math:`x`.

    If :math:`n \neq 1`, then the vectorization is repeated :math:`n` times.
    """
    if hasattr(x, "diag") and n == 1:
        return x.diag
    elif hasattr(x, "dupdiag"):
        return x.dupdiag(n)


@_error_on_none
@convert_and_refine_arguments("x")
def maindiag(x):
    """Extract the diagonal of :math:`x` as a column vector."""
    if hasattr(x, "maindiag"):
        return x.maindiag


@_error_on_none
@convert_and_refine_arguments("x")
def trace(x):
    """Denote the trace of a square matrix."""
    if hasattr(x, "tr"):
        return x.tr


@_error_on_none
@convert_and_refine_arguments("x")
def partial_trace(x, subsystems=0, dimensions=2, k=None, dim=None):
    """See :meth:`.expressions.ComplexAffineExpression.partial_trace`.

    The parameters `k` and `dim` are for backwards compatibility.
    """
    if k is not None:
        throw_deprecation_warning("Argument 'k' to partial_trace is "
            "deprecated: Use 'subsystems' instead.", decoratorLevel=2)
        subsystems = k

    if dim is not None:
        throw_deprecation_warning("Argument 'dim' to partial_trace is "
            "deprecated: Use 'dimensions' instead.", decoratorLevel=2)
        dimensions = dim

    if isinstance(x, ComplexAffineExpression):
        return x.partial_trace(subsystems, dimensions)


@_error_on_none
@convert_and_refine_arguments("x")
def partial_transpose(x, subsystems=0, dimensions=2, k=None, dim=None):
    """See :meth:`.expressions.ComplexAffineExpression.partial_transpose`.

    The parameters `k` and `dim` are for backwards compatibility.
    """
    if k is not None:
        throw_deprecation_warning("Argument 'k' to partial_transpose is "
            "deprecated: Use 'subsystems' instead.", decoratorLevel=2)
        subsystems = k

    if dim is not None:
        throw_deprecation_warning("Argument 'dim' to partial_transpose is "
            "deprecated: Use 'dimensions' instead.", decoratorLevel=2)
        dimensions = dim

    if isinstance(x, ComplexAffineExpression):
        return x.partial_transpose(subsystems, dimensions)


# ------------------------------------------------------------------------------
# Alias functions for expression classes meant to be instanciated by the user.
# ------------------------------------------------------------------------------


def _shorthand(cls):
    def shorthand(*args, **kwargs):
        return cls(*args, **kwargs)

    shorthand.__doc__ = "Shorthand for :class:`~picos.{}`.".format(cls.__name__)

    return shorthand


expcone = _shorthand(ExponentialCone)
geomean = _shorthand(GeometricMean)
kldiv   = _shorthand(NegativeEntropy)
lse     = _shorthand(LogSumExp)
rsoc    = _shorthand(RotatedSecondOrderCone)
soc     = _shorthand(SecondOrderCone)
sumexp  = _shorthand(SumExponentials)


def max(x):
    """Wrapper for :class:`~picos.SumExtremes`.

    Sets ``k = 1``, ``largest = True`` and ``eigenvalues = False``.

    :Example:

    >>> from picos import RealVariable, max
    >>> x = RealVariable("x", 5)
    >>> max(x)
    <Largest Element: max(x)>
    >>> max(x) <= 2  # The same as x <= 2.
    <Largest Element Constraint: max(x) ≤ 2>
    """
    return SumExtremes(x, 1, largest=True, eigenvalues=False)


def min(x):
    """Wrapper for :class:`~picos.SumExtremes`.

    Sets ``k = 1``, ``largest = False`` and ``eigenvalues = False``.

    :Example:

    >>> from picos import RealVariable, min
    >>> x = RealVariable("x", 5)
    >>> min(x)
    <Smallest Element: min(x)>
    >>> min(x) >= 2  # The same as x >= 2.
    <Smallest Element Constraint: min(x) ≥ 2>
    """
    return SumExtremes(x, 1, largest=False, eigenvalues=False)


def sum_k_largest(x, k):
    """Wrapper for :class:`~picos.SumExtremes`.

    Sets ``largest = True`` and ``eigenvalues = False``.

    :Example:

    >>> from picos import RealVariable, sum_k_largest
    >>> x = RealVariable("x", 5)
    >>> sum_k_largest(x, 2)
    <Sum of Largest Elements: sum_2_largest(x)>
    >>> sum_k_largest(x, 2) <= 2
    <Sum of Largest Elements Constraint: sum_2_largest(x) ≤ 2>
    """
    return SumExtremes(x, k, largest=True, eigenvalues=False)


def sum_k_smallest(x, k):
    """Wrapper for :class:`~picos.SumExtremes`.

    Sets ``largest = False`` and ``eigenvalues = False``.

    :Example:

    >>> from picos import RealVariable, sum_k_smallest
    >>> x = RealVariable("x", 5)
    >>> sum_k_smallest(x, 2)
    <Sum of Smallest Elements: sum_2_smallest(x)>
    >>> sum_k_smallest(x, 2) >= 2
    <Sum of Smallest Elements Constraint: sum_2_smallest(x) ≥ 2>
    """
    return SumExtremes(x, k, largest=False, eigenvalues=False)


def lambda_max(x):
    """Wrapper for :class:`~picos.SumExtremes`.

    Sets ``k = 1``, ``largest = True`` and ``eigenvalues = True``.

    :Example:

    >>> from picos import SymmetricVariable, lambda_max
    >>> X = SymmetricVariable("X", 5)
    >>> lambda_max(X)
    <Largest Eigenvalue: λ_max(X)>
    >>> lambda_max(X) <= 2
    <Largest Eigenvalue Constraint: λ_max(X) ≤ 2>
    """
    return SumExtremes(x, 1, largest=True, eigenvalues=True)


def lambda_min(x):
    """Wrapper for :class:`~picos.SumExtremes`.

    Sets ``k = 1``, ``largest = False`` and ``eigenvalues = True``.

    :Example:

    >>> from picos import SymmetricVariable, lambda_min
    >>> X = SymmetricVariable("X", 5)
    >>> lambda_min(X)
    <Smallest Eigenvalue: λ_min(X)>
    >>> lambda_min(X) >= 2
    <Smallest Eigenvalue Constraint: λ_min(X) ≥ 2>
    """
    return SumExtremes(x, 1, largest=False, eigenvalues=True)


def sum_k_largest_lambda(x, k):
    """Wrapper for :class:`~picos.SumExtremes`.

    Sets ``largest = True`` and ``eigenvalues = True``.

    :Example:

    >>> from picos import SymmetricVariable, sum_k_largest_lambda
    >>> X = SymmetricVariable("X", 5)
    >>> sum_k_largest_lambda(X, 2)
    <Sum of Largest Eigenvalues: sum_2_largest_λ(X)>
    >>> sum_k_largest_lambda(X, 2) <= 2
    <Sum of Largest Eigenvalues Constraint: sum_2_largest_λ(X) ≤ 2>
    """
    return SumExtremes(x, k, largest=True, eigenvalues=True)


def sum_k_smallest_lambda(x, k):
    """Wrapper for :class:`~picos.SumExtremes`.

    Sets ``largest = False`` and ``eigenvalues = True``.

    :Example:

    >>> from picos import SymmetricVariable, sum_k_smallest_lambda
    >>> X = SymmetricVariable("X", 5)
    >>> sum_k_smallest_lambda(X, 2)
    <Sum of Smallest Eigenvalues: sum_2_smallest_λ(X)>
    >>> sum_k_smallest_lambda(X, 2) >= 2
    <Sum of Smallest Eigenvalues Constraint: sum_2_smallest_λ(X) ≥ 2>
    """
    return SumExtremes(x, k, largest=False, eigenvalues=True)


# ------------------------------------------------------------------------------
# Legacy algebraic functions for backwards compatibility.
# ------------------------------------------------------------------------------


def _deprecated_shorthand(cls, new_shorthand=None):
    shRef = "~picos.{}".format(cls.__name__)
    uiRef = "~picos.{}".format(new_shorthand) if new_shorthand else shRef

    # FIXME: This doesn't show the name of the deprecated shorthand function.
    @deprecated("2.0", useInstead=uiRef)
    def shorthand(*args, **kwargs):
        """|PLACEHOLDER|"""  # noqa
        return cls(*args, **kwargs)

    shorthand.__doc__ = shorthand.__doc__.replace("|PLACEHOLDER|",
        "Legacy shorthand for :class:`{}`.".format(shRef))

    return shorthand


ball             = _deprecated_shorthand(Ball)
detrootn         = _deprecated_shorthand(DetRootN)
kullback_leibler = _deprecated_shorthand(NegativeEntropy, "kldiv")
logsumexp        = _deprecated_shorthand(LogSumExp, "lse")
norm             = _deprecated_shorthand(Norm)


@deprecated("2.0", useInstead="~picos.PowerTrace")
def tracepow(exp, num=1, denom=1, coef=None):
    """Legacy shorthand for :class:`~picos.PowerTrace`."""
    return PowerTrace(exp, num / denom, coef)


@deprecated("2.0", useInstead="~picos.Constant")
def new_param(name, value):
    """Create a constant or a list or dict or tuple thereof."""
    if isinstance(value, list):
        # Handle a vector.
        try:
            for x in value:
                complex(x)
        except Exception:
            pass
        else:
            return Constant(name, value)

        # Handle a matrix.
        if all(isinstance(x, list) for x in value) \
        and all(len(x) == len(value[0]) for x in value):
            try:
                for x in value:
                    for y in x:
                        complex(y)
            except Exception:
                pass
            else:
                return Constant(name, value)

        # Return a list of constants.
        return [Constant(glyphs.slice(name, i), x) for i, x in enumerate(value)]
    elif isinstance(value, tuple):
        # Return a list of constants.
        # NOTE: This is very inconsistent, but legacy behavior.
        return [Constant(glyphs.slice(name, i), x) for i, x in enumerate(value)]
    elif isinstance(value, dict):
        return {k: Constant(glyphs.slice(name, k), x) for k, x in value.items()}
    else:
        return Constant(name, value)


@deprecated("2.0", useInstead="~picos.FlowConstraint")
def flow_Constraint(*args, **kwargs):
    """Legacy shorthand for :class:`~picos.FlowConstraint`."""
    from ..constraints.con_flow import FlowConstraint
    return FlowConstraint(*args, **kwargs)


@deprecated("2.0", useInstead="~picos.maindiag")
def diag_vect(x):
    """Extract the diagonal of :math:`x` as a column vector."""
    return maindiag(x)


@deprecated("2.0", useInstead="~picos.Simplex")
def simplex(gamma):
    r"""Create a standard simplex of radius :math:`\gamma`."""
    return Simplex(gamma, truncated=False, symmetrized=False)


@deprecated("2.0", useInstead="~picos.Simplex")
def truncated_simplex(gamma, sym=False):
    r"""Create a truncated simplex of radius :math:`\gamma`."""
    return Simplex(gamma, truncated=True, symmetrized=sym)


# --------------------------------------
__all__ = api_end(_API_START, globals())



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2018-2019 Maximilian Stahlberg
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Implementation of :class:`LogConstraint`."""

from collections import namedtuple

from .. import glyphs
from ..apidoc import api_end, api_start
from .constraint import Constraint, ConstraintConversion

_API_START = api_start(globals())
# -------------------------------


class LogConstraint(Constraint):
    """Lower bound on a logarithm."""

    class ExpConeConversion(ConstraintConversion):
        """Bound on a logarithm to exponential cone constraint conversion."""

        @classmethod
        def predict(cls, subtype, options):
            """Implement :meth:`~.constraint.ConstraintConversion.predict`."""
            from . import AffineConstraint, ExpConeConstraint

            yield ("con", AffineConstraint.make_type(dim=1, eq=False), 1)
            yield ("con", ExpConeConstraint.make_type(), 1)

        @classmethod
        def convert(cls, con, options):
            """Implement :meth:`~.constraint.ConstraintConversion.convert`."""
            from ..expressions import ExponentialCone
            from ..modeling import Problem

            x = con.log.x
            t = con.lb

            P = Problem()
            P.add_constraint(x >= 0)
            P.add_constraint((x // 1 // t) << ExponentialCone())
            return P

    def __init__(self, log, lowerBound):
        """Construct a :class:`LogConstraint`.

        :param ~picos.expressions.Logarithm log:
            Constrained expression.
        :param ~picos.expressions.AffineExpression lowerBound:
            Lower bound on the expression.
        """
        from ..expressions import AffineExpression, Logarithm

        assert isinstance(log, Logarithm)
        assert isinstance(lowerBound, AffineExpression)
        assert len(lowerBound) == 1

        self.log = log
        self.lb  = lowerBound

        super(LogConstraint, self).__init__("Logarithmic")

    Subtype = namedtuple("Subtype", ())

    def _subtype(self):
        return self.Subtype()

    @classmethod
    def _cost(cls, subtype):
        return 2

    def _expression_names(self):
        yield "log"
        yield "lb"

    def _str(self):
        return glyphs.ge(self.log.string, self.lb.string)

    def _get_slack(self):
        return self.log.value - self.lb.value


# --------------------------------------
__all__ = api_end(_API_START, globals())



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2017-2019 Maximilian Stahlberg
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Implementation of :class:`GLPKSolver`."""

import cvxopt

from ..apidoc import api_end, api_start
from ..compat import assert_import_exists
from ..constraints import AffineConstraint
from ..expressions import AffineExpression, BinaryVariable, IntegerVariable
from ..modeling.footprint import Specification
from ..modeling.solution import (PS_FEASIBLE, PS_INFEASIBLE, PS_UNBOUNDED,
                                 PS_UNKNOWN, SS_EMPTY, SS_FEASIBLE,
                                 SS_INFEASIBLE, SS_OPTIMAL, SS_UNKNOWN)
from .solver import Solver

_API_START = api_start(globals())
# -------------------------------


class GLPKSolver(Solver):
    """Interface to the GLPK solver via swiglpk."""

    SUPPORTED = Specification.compile(
        directions=Specification.ALL,
        objectives=[
            AffineExpression],
        variables=Specification.ALL,
        constraints=[
            AffineConstraint])

    @classmethod
    def supports(cls, footprint):
        """Implement :meth:`~.solver.Solver.supports`."""
        if not Solver.supports(footprint):
            return False

        return footprint << cls.SUPPORTED

    @classmethod
    def default_penalty(cls):
        """Implement :meth:`~.solver.Solver.default_penalty`."""
        return 1.0  # Stable free/open source solver.

    @classmethod
    def test_availability(cls):
        """Implement :meth:`~.solver.Solver.test_availability`."""
        assert_import_exists("swiglpk")

    @classmethod
    def names(cls):
        """Implement :meth:`~.solver.Solver.names`."""
        return "glpk", "GLPK", "GNU Linear Programming Kit"

    def __init__(self, problem):
        """Initialize a GLPK solver interface.

        :param ~picos.Problem problem: The problem to be solved.
        """
        self._glpkVarOffset = {}
        self._glpkConOffset = {}

        super(GLPKSolver, self).__init__(problem)

    def __del__(self):
        try:
            import swiglpk as glpk
        except ImportError:
            # Happens when python is shutting down. In this case, no garbage
            # collection is necessary, anyway.
            return

        if self.int is not None:
            glpk.glp_delete_prob(self.int)

    def reset_problem(self):
        """Implement :meth:`~.solver.Solver.reset_problem`."""
        import swiglpk as glpk

        self._glpkVarOffset.clear()
        self._glpkConOffset.clear()

        if self.int is not None:
            glpk.glp_delete_prob(self.int)
            self.int = None

    def _import_variable(self, picosVar):
        import swiglpk as glpk

        p = self.int
        dim = picosVar.dim

        # Append columns to the constraint matrix.
        offset = glpk.glp_add_cols(p, dim)

        # Retrieve the type.
        if isinstance(picosVar, IntegerVariable):
            varType = glpk.GLP_IV
        elif isinstance(picosVar, BinaryVariable):
            varType = glpk.GLP_BV
        else:
            varType = glpk.GLP_CV

        # Retrieve bounds.
        if not isinstance(picosVar, BinaryVariable):
            boundKeys   = [glpk.GLP_FR]*dim
            lowerBounds = [0]*dim
            upperBounds = [0]*dim

            lower, upper = picosVar.bound_dicts

            for i, b in lower.items():
                boundKeys[i] = glpk.GLP_LO
                lowerBounds[i] = b

            for i, b in upper.items():
                if boundKeys[i] == glpk.GLP_FR:
                    boundKeys[i] = glpk.GLP_UP
                else:  # Also has a lower bound.
                    if lowerBounds[i] == b:
                        boundKeys[i] = glpk.GLP_FX
                    else:
                        boundKeys[i] = glpk.GLP_DB

                upperBounds[i] = b

        # Import scalar variables.
        for i in range(picosVar.dim):
            glpkIndex = offset + i

            # Assign a name.
            glpk.glp_set_col_name(
                p, glpkIndex, "{}_{}".format(picosVar.name, i))

            # Set the type.
            glpk.glp_set_col_kind(p, glpkIndex, varType)

            # Assign the bounds.
            if not isinstance(picosVar, BinaryVariable):
                glpk.glp_set_col_bnds(
                    p, glpkIndex, boundKeys[i], lowerBounds[i], upperBounds[i])

        self._glpkVarOffset[picosVar] = offset

    def _remove_variable(self, picosVar):
        import swiglpk as glpk

        dim = picosVar.dim

        offset = self._glpkVarOffset.pop(picosVar)

        glpkVars = glpk.intArray(dim + 1)
        for i in range(dim):
            glpkVars[i + 1] = offset + i  # Index 0 is unused.

        glpk.glp_del_cols(self.int, dim, glpkVars)

        for otherVar, otherOffset in self._glpkVarOffset.items():
            if otherOffset > offset:
                self._glpkVarOffset[otherVar] -= dim

    def _import_constraint(self, picosCon):
        import swiglpk as glpk

        p = self.int

        # Append rows to the constraint matrix.
        rowOffset = glpk.glp_add_rows(p, len(picosCon))

        # Import scalar constraints.
        rows = picosCon.sparse_Ab_rows(self._glpkVarOffset)
        for localConIndex, (glpkVarIndices, coefs, rhs) in enumerate(rows):
            glpkConIndex = rowOffset + localConIndex
            numColumns   = len(glpkVarIndices)

            # Assign a name.
            glpk.glp_set_row_name(
                p, glpkConIndex, "{}_{}".format(picosCon.id, localConIndex))

            # Set the constant term.
            if picosCon.is_equality():
                glpk.glp_set_row_bnds(p, glpkConIndex, glpk.GLP_FX, rhs, rhs)
            elif picosCon.is_increasing():
                glpk.glp_set_row_bnds(p, glpkConIndex, glpk.GLP_UP, 0, rhs)
            elif picosCon.is_decreasing():
                glpk.glp_set_row_bnds(p, glpkConIndex, glpk.GLP_LO, rhs, 0)
            else:
                assert False, "Unexpected constraint relation."

            # Set coefficients.
            # NOTE: GLPK requires a glpk.intArray containing column indices and
            #       a glpk.doubleArray of same size containing the coefficients
            #       for the listed column index. The first element of both
            #       arrays (with index 0) is skipped by GLPK.
            glpkVarIndicesArray = glpk.intArray(numColumns + 1)
            for i in range(numColumns):
                glpkVarIndicesArray[i + 1] = glpkVarIndices[i]

            coefficientsArray = glpk.doubleArray(numColumns + 1)
            for i in range(numColumns):
                coefficientsArray[i + 1] = coefs[i]

            glpk.glp_set_mat_row(p, glpkConIndex, numColumns,
                glpkVarIndicesArray, coefficientsArray)

        self._glpkConOffset[picosCon] = rowOffset

    def _remove_constraint(self, picosCon):
        import swiglpk as glpk

        length = len(picosCon)

        offset = self._glpkConOffset.pop(picosCon)

        glpkCons = glpk.intArray(length + 1)
        for i in range(length):
            glpkCons[i + 1] = offset + i  # Index 0 is unused.

        glpk.glp_del_rows(self.int, length, glpkCons)

        for otherCon, otherOffset in self._glpkConOffset.items():
            if otherOffset > offset:
                self._glpkConOffset[otherCon] -= length

    def _import_objective(self):
        import swiglpk as glpk

        p = self.int
        direction, objective = self.ext.no

        # Set optimization direction.
        if direction == "min":
            glpk.glp_set_obj_dir(p, glpk.GLP_MIN)
        else:
            assert direction == "max"
            glpk.glp_set_obj_dir(p, glpk.GLP_MAX)

        # Set objective function shift (index 0).
        glpk.glp_set_obj_coef(p, 0, objective._const[0])

        # Set objective function coefficient of the scalar variable.
        for picosVar, picosCoef in objective._coefs.items():
            for localIndex in range(picosVar.dim):
                if picosCoef[localIndex]:
                    glpkIndex = self._glpkVarOffset[picosVar] + localIndex
                    glpk.glp_set_obj_coef(p, glpkIndex, picosCoef[localIndex])

    def _reset_objective(self):
        import swiglpk as glpk

        p = self.int

        # Zero the objective.
        scalarVars = glpk.glp_get_num_cols(p)
        for i in range(scalarVars + 1):  # Index 0 refers to the constant term.
            glpk.glp_set_obj_coef(p, i, 0)

    def _import_problem(self):
        import swiglpk as glpk

        if self.verbosity() >= 1:
            glpk.glp_term_out(glpk.GLP_ON)
        else:
            glpk.glp_term_out(glpk.GLP_OFF)

        # Create a problem instance.
        self.int = glpk.glp_create_prob()

        # Import variables.
        for variable in self.ext.variables.values():
            self._import_variable(variable)

        # Import constraints.
        for constraint in self.ext.constraints.values():
            self._import_constraint(constraint)

        # Set objective.
        self._import_objective()

    def _update_problem(self):
        import swiglpk as glpk

        resetBasis = False

        for oldConstraint in self._removed_constraints():
            self._remove_constraint(oldConstraint)
            resetBasis = True

        for oldVariable in self._removed_variables():
            self._remove_variable(oldVariable)
            resetBasis = True

        for newVariable in self._new_variables():
            self._import_variable(newVariable)

        for newConstraint in self._new_constraints():
            self._import_constraint(newConstraint)

        if self._objective_has_changed():
            self._reset_objective()
            self._import_objective()

        if resetBasis:
            # TODO: Repair the basis in _remove_constraint, _remove_variable.
            glpk.glp_cpx_basis(self.int)

    def _solve(self):
        import swiglpk as glpk

        p = self.int

        continuous = self.ext.is_continuous()
        minimizing = glpk.glp_get_obj_dir(p) == glpk.GLP_MIN

        # Select LP solver (Simplex or Interior Point Method).
        if continuous:
            if self.ext.options.lp_root_method == "interior":
                interior = True
            else:
                # Default to Simplex.
                interior = False
            simplex = not interior
        else:
            simplex = interior = False

        # Select appropriate options container.
        if simplex:
            options = glpk.glp_smcp()
            glpk.glp_init_smcp(options)
        elif interior:
            options = glpk.glp_iptcp()
            glpk.glp_init_iptcp(options)
        else:
            options = glpk.glp_iocp()
            glpk.glp_init_iocp(options)

        # verbosity
        verbosity = self.verbosity()
        if verbosity < 0:
            options.msg_lev = glpk.GLP_MSG_OFF
        elif verbosity == 0:
            options.msg_lev = glpk.GLP_MSG_ERR
        elif verbosity == 1:
            options.msg_lev = glpk.GLP_MSG_ON
        elif verbosity >= 2:
            options.msg_lev = glpk.GLP_MSG_ALL

        # abs_prim_fsb_tol
        if self.ext.options.abs_prim_fsb_tol is not None:
            options.tol_bnd = self.ext.options.abs_prim_fsb_tol

        # abs_dual_fsb_tol
        if self.ext.options.abs_dual_fsb_tol is not None:
            options.tol_dj = self.ext.options.abs_dual_fsb_tol

        # rel_bnb_opt_tol
        # Note that the option is silently ignored if passed alongside an LP;
        # while the solver does not allow us to pass the option in that case, it
        # is still technically a valid option as every LP is also a MIP.
        if self.ext.options.rel_bnb_opt_tol is not None:
            if not continuous:
                options.mip_gap = self.ext.options.rel_bnb_opt_tol

        # max_iterations
        if not simplex:
            self._handle_unsupported_option("max_iterations",
                "GLPK supports the 'max_iterations' option only with Simplex.")
        elif self.ext.options.max_iterations is not None:
            options.it_lim = int(self.ext.options.max_iterations)

        # lp_root_method
        # Note that the PICOS option is explicitly also meant for the MIP
        # preprocessing step but GLPK does not support it in that scenario.
        if not continuous:
            self._handle_unsupported_option("lp_root_method",
                "GLPK supports the 'lp_root_method' option only for LPs.")
        elif self.ext.options.lp_root_method is not None:
            if self.ext.options.lp_root_method == "interior":
                # Handled above.
                pass
            elif self.ext.options.lp_root_method == "psimplex":
                assert simplex
                options.meth = glpk.GLP_PRIMAL
            elif self.ext.options.lp_root_method == "dsimplex":
                assert simplex
                options.meth = glpk.GLP_DUAL
            else:
                assert False, "Unexpected lp_root_method value."

        # timelimit
        if interior:
            self._handle_unsupported_option("timelimit",
                "GLPK does not support the 'timelimit' option with the "
                "Interior Point Method.")
        elif self.ext.options.timelimit is not None:
            options.tm_lim = int(1000 * self.ext.options.timelimit)

        # Handle unsupported options.
        self._handle_unsupported_options(
            "lp_node_method", "treememory", "max_fsb_nodes", "hotstart")

        # TODO: Add GLPK-sepcific options. Candidates are:
        #       For both Simplex and MIPs:
        #           tol_*, out_*
        #       For Simplex:
        #           pricing, r_test, obj_*
        #       For the Interior Point Method:
        #           ord_alg
        #       For MIPs:
        #           *_tech, *_heur, ps_tm_lim, *_cuts, cb_size, binarize

        # Attempt to solve the problem.
        with self._header():
            with self._stopwatch():
                if simplex:
                    # TODO: Support glp_exact.
                    error = glpk.glp_simplex(p, options)
                elif interior:
                    error = glpk.glp_interior(p, options)
                else:
                    options.presolve = glpk.GLP_ON
                    error = glpk.glp_intopt(p, options)

            # Conert error codes to text output.
            # Note that by printing it above the footer, this output is made to
            # look like it's coming from GLPK, which is technically wrong but
            # semantically correct.
            if error == glpk.GLP_EBADB:
                self._warn("Unable to start the search, because the initial "
                    "basis specified in the problem object is invalid.")
            elif error == glpk.GLP_ESING:
                self._warn("Unable to start the search, because the basis "
                    "matrix corresponding to the initial basis is singular "
                    "within the working precision.")
            elif error == glpk.GLP_ECOND:
                self._warn("Unable to start the search, because the basis "
                    "matrix corresponding to the initial basis is "
                    "ill-conditioned.")
            elif error == glpk.GLP_EBOUND:
                self._warn("Unable to start the search, because some double-"
                    "bounded variables have incorrect bounds.")
            elif error == glpk.GLP_EFAIL:
                self._warn("The search was prematurely terminated due to a "
                    "solver failure.")
            elif error == glpk.GLP_EOBJLL:
                self._warn("The search was prematurely terminated, because the "
                    "objective function being maximized has reached its lower "
                    "limit and continues decreasing.")
            elif error == glpk.GLP_EOBJUL:
                self._warn("The search was prematurely terminated, because the "
                    "objective function being minimized has reached its upper "
                    "limit and continues increasing.")
            elif error == glpk.GLP_EITLIM:
                self._warn("The search was prematurely terminated, because the "
                    "simplex iteration limit has been exceeded.")
            elif error == glpk.GLP_ETMLIM:
                self._warn("The search was prematurely terminated, because the "
                    "time limit has been exceeded.")
            elif error == glpk.GLP_ENOPFS:
                self._verbose("The LP has no primal feasible solution.")
            elif error == glpk.GLP_ENODFS:
                self._verbose("The LP has no dual feasible solution.")
            elif error != 0:
                self._warn("GLPK error {:d}.".format(error))

        # Retrieve primals.
        primals = {}
        if self.ext.options.primals is not False:
            for variable in self.ext.variables.values():
                value = []

                for localIndex in range(variable.dim):
                    glpkIndex = self._glpkVarOffset[variable] + localIndex

                    if simplex:
                        localValue = glpk.glp_get_col_prim(p, glpkIndex)
                    elif interior:
                        localValue = glpk.glp_ipt_col_prim(p, glpkIndex)
                    else:
                        localValue = glpk.glp_mip_col_val(p, glpkIndex)

                    value.append(localValue)

                primals[variable] = value

        # Retrieve duals.
        duals = {}
        if self.ext.options.duals is not False and continuous:
            for constraint in self.ext.constraints.values():
                value = []

                for localIndex in range(len(constraint)):
                    glpkIndex = self._glpkConOffset[constraint] + localIndex

                    if simplex:
                        localValue = glpk.glp_get_row_dual(p, glpkIndex)
                    elif interior:
                        localValue = glpk.glp_ipt_row_dual(p, glpkIndex)
                    else:
                        assert False

                    value.append(localValue)

                dual = cvxopt.matrix(value, constraint.size)
                if constraint.is_decreasing() ^ minimizing:
                    dual = -dual

                duals[constraint] = dual

        # # Retrieve objective value.
        # if simplex:
        #     objectiveValue = glpk.glp_get_obj_val(p)
        # elif interior:
        #     objectiveValue = glpk.glp_ipt_obj_val(p)
        # else:
        #     objectiveValue = glpk.glp_mip_obj_val(p)

        if simplex:
            probStatusCode = glpk.glp_get_status(p)
        elif interior:
            probStatusCode = glpk.glp_ipt_status(p)
        else:
            probStatusCode = glpk.glp_mip_status(p)

        if probStatusCode == glpk.GLP_OPT:
            # simplex, interior, mip
            problemStatus = PS_FEASIBLE
        elif probStatusCode == glpk.GLP_FEAS:
            # simplex, mip
            problemStatus = PS_FEASIBLE
        elif probStatusCode == glpk.GLP_INFEAS:
            # simplex, interior
            problemStatus = PS_UNKNOWN
        elif probStatusCode == glpk.GLP_NOFEAS:
            # simplex, interior, mip
            problemStatus = PS_INFEASIBLE
        elif probStatusCode == glpk.GLP_UNBND:
            # simplex
            problemStatus = PS_UNBOUNDED
        elif probStatusCode == glpk.GLP_UNDEF:
            # simplex, interior, mip
            problemStatus = PS_UNKNOWN
        else:
            problemStatus = PS_UNKNOWN

        if simplex:
            prmlStatusCode = glpk.glp_get_prim_stat(p)
            dualStatusCode = glpk.glp_get_dual_stat(p)

            if prmlStatusCode == glpk.GLP_FEAS:
                if probStatusCode == glpk.GLP_OPT:
                    prmlStatus = SS_OPTIMAL
                else:
                    prmlStatus = SS_FEASIBLE
            elif prmlStatusCode == glpk.GLP_INFEAS:
                prmlStatus = SS_INFEASIBLE
            elif prmlStatusCode == glpk.GLP_NOFEAS:
                prmlStatus = PS_INFEASIBLE
            elif prmlStatusCode == glpk.GLP_UNDEF:
                prmlStatus = SS_EMPTY
            else:
                prmlStatus = SS_UNKNOWN

            if dualStatusCode == glpk.GLP_FEAS:
                if probStatusCode == glpk.GLP_OPT:
                    dualStatus = SS_OPTIMAL
                else:
                    dualStatus = SS_FEASIBLE
            elif dualStatusCode == glpk.GLP_INFEAS:
                dualStatus = SS_INFEASIBLE
            elif dualStatusCode == glpk.GLP_NOFEAS:
                dualStatus = PS_INFEASIBLE
            elif dualStatusCode == glpk.GLP_UNDEF:
                dualStatus = SS_EMPTY
            else:
                dualStatus = SS_UNKNOWN
        elif interior:
            if probStatusCode == glpk.GLP_UNDEF:
                prmlStatus = SS_EMPTY
                dualStatus = SS_EMPTY
            elif probStatusCode == glpk.GLP_OPT:
                prmlStatus = SS_OPTIMAL
                dualStatus = SS_OPTIMAL
            elif probStatusCode == glpk.GLP_FEAS:
                prmlStatus = SS_FEASIBLE
                dualStatus = SS_FEASIBLE
            elif probStatusCode == glpk.GLP_NOFEAS:
                prmlStatus = SS_INFEASIBLE
                dualStatus = SS_INFEASIBLE
            else:
                prmlStatus = SS_UNKNOWN
                dualStatus = SS_UNKNOWN
        else:  # MIP
            if probStatusCode == glpk.GLP_UNDEF:
                prmlStatus = SS_EMPTY
            elif probStatusCode == glpk.GLP_OPT:
                prmlStatus = SS_OPTIMAL
            elif probStatusCode == glpk.GLP_FEAS:
                prmlStatus = SS_FEASIBLE
            elif probStatusCode == glpk.GLP_NOFEAS:
                prmlStatus = SS_INFEASIBLE
            else:
                prmlStatus = SS_UNKNOWN

            dualStatus = SS_EMPTY

        return self._make_solution(
            primals, duals, prmlStatus, dualStatus, problemStatus)


# --------------------------------------
__all__ = api_end(_API_START, globals())



# coding: utf-8

# ------------------------------------------------------------------------------
# Copyright (C) 2019 Maximilian Stahlberg, 2020 Guillaume Sagnol
# Based on the original picos.expressions module by Guillaume Sagnol.
#
# This file is part of PICOS.
#
# PICOS is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PICOS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------

"""Implements :class:`Norm`."""

import operator
from collections import namedtuple

import cvxopt
import numpy

from .. import glyphs
from ..apidoc import api_end, api_start
from ..caching import cached_unary_operator
from ..constraints import (AbsoluteValueConstraint, Constraint,
                           MatrixNormConstraint, SOCConstraint,
                           VectorNormConstraint, SpectralNormConstraint,
                           NuclearNormConstraint)
from ..legacy import throw_deprecation_warning
from .data import (convert_and_refine_arguments, convert_operands, cvx2np,
                   make_fraction)
from .exp_affine import AffineExpression, ComplexAffineExpression
from .expression import Expression, refine_operands, validate_prediction

_API_START = api_start(globals())
# -------------------------------


class Norm(Expression):
    r"""Entrywise :math:`p`-norm or :math:`L_{p,q}`-norm of an expression.

    This class can represent the absolute value, the modulus, a vector
    :math:`p`-norm and an entrywise matrix :math:`L_{p,q}`-norm of a (complex)
    affine expression. In addition to these convex norms, it can represent the
    concave *generalized vector* :math:`p`-*norm* with :math:`0 < p \leq 1`.

    Not all of these norms are available on the complex field; see the
    definitions below to learn more.

    :Definition:

    If :math:`q` is not given (:obj:`None`), then it is set equal to :math:`p`.

    1.  If the normed expression is a real scalar :math:`x`, then this is the
        absolute value

        .. math::

            |x| = \begin{cases}
               x, &\text{if }x \geq 0,\\
               -x, &\text{otherwise.}
            \end{cases}

        The parameters :math:`p` and :math:`q` are ignored in this case.

    2.  If the normed expression is a complex scalar :math:`z`, then this is the
        modulus

        .. math::

            |z| = \sqrt{\operatorname{Re}(z)^2 + \operatorname{Im}(z)^2}.

        The parameters :math:`p` and :math:`q` are ignored in this case.

    3.  If the normed expression is a real vector :math:`x` and :math:`p = q`,
        then this is the (generalized) vector :math:`p`-norm

        .. math::

            \lVert x \rVert_p = \left(\sum_{i=1}^n |x_i|^p\right)^{\frac{1}{p}}

        for :math:`p \in \mathbb{Q} \cup \{\infty\}` with :math:`p > 0` and
        :math:`|x_i|` the absolute value of the :math:`i`-th entry of :math:`x`.

        Note that for :math:`p < 1` the expression is not convex and thus not a
        proper norm. However, it is concave over the nonnegative orthant and
        posing a lower bound on such a generalized norm yields a convex
        constraint for :math:`x \geq 0`.

        .. warning::

            When you pose a lower bound on a concave generalized norm
            (:math:`p < 1`), then PICOS enforces :math:`x \geq 0` through an
            auxiliary constraint during solution search.

        Special cases:

        -   For :math:`p = 1`, this is the *Manhattan* or *Taxicab* norm
            :math:`\lVert x \rVert_{\text{sum}}`.
        -   For :math:`p = 2`, this is the Euclidean norm
            :math:`\lVert x \rVert = \lVert x \rVert_2`.
        -   For :math:`p = \infty`, this is the *Maximum*, *Chebyshev*, or
            *Infinity* norm :math:`\lVert x \rVert_{\text{max}}`.

    4.  If the normed expression is a real vector :math:`x` and
        :math:`p \neq q`, then it is treated as a matrix with a single row or a
        single column, depending on the shape associated with :math:`x`. See
        case (5).

    5.  If the normed expression is a complex vector :math:`z` and
        :math:`p = q`, then the definition is the same as in case (3) but with
        :math:`x = z`, :math:`|x_i|` the modulus of the :math:`i`-th entry of
        :math:`x`, and :math:`p \geq 1`.

    6.  If the normed expression is a real :math:`m \times n` matrix :math:`X`,
        then this is the :math:`L_{p,q}`-norm

        .. math::

            \lVert X \rVert_{p,q} =
            \left(\sum_{j = 1}^n \left(\sum_{i = 1}^m
                |X_{ij}|^p
            \right)^{\frac{q}{p}} \right)^{\frac{1}{q}}

        for :math:`p, q \in \mathbb{Q} \cup \{\infty\}` with :math:`p,q \geq 1`.

        If :math:`p = q`, then this is equal to the (generalized) vector
        :math:`p`-norm of the the vectorized matrix, that is
        :math:`\lVert X \rVert_{p,p} = \lVert \operatorname{vec}(X) \rVert_p`.
        In this case, the requirement :math:`p \geq 1` is relaxed to
        :math:`p > 0` and :math:`X` may be a complex matrix. See case (3).

        Special cases:

        -   For :math:`p = q = 2`, this is the Frobenius norm
            :math:`\lVert X \rVert = \lVert X \rVert_F`.
        -   For :math:`p = 1`, :math:`q = \infty`, this is the maximum absolute
            column sum

            .. math::

                \lVert X \rVert_{1,\infty} =
                \max_{j=1 \ldots n} \sum_{i=1}^m |X_{ij}|.

            This equals the operator norm induced by the vector :math:`1`-norm.
            You can obtain the maximum absolute row sum (the operator norm
            induced by the vector :math:`\infty`-norm) by first transposing
            :math:`X`.

    7.  Complex matrix norms are not supported.

    .. note::

        You can write :math:`\infty` in Python as ``float("inf")``.
    """

    # --------------------------------------------------------------------------
    # Initialization and factory methods.
    # --------------------------------------------------------------------------

    @convert_and_refine_arguments("x")
    def __init__(self, x, p=2, q=None, denominator_limit=1000):
        """Construct a :class:`Norm`.

        :param x: The affine expression to take the norm of.
        :type x: ~picos.expressions.ComplexAffineExpression
        :param float p: The value for :math:`p`, which is cast to a limited
            precision fraction.
        :param float q: The value for :math:`q`, which is cast to a limited
            precision fraction. The default of :obj:`None` means *equal to*
            :math:`p`.
        :param int denominator_limit: The largest allowed denominator when
            casting :math:`p` and :math:`q` to a fraction. Higher values can
            yield a greater precision at reduced performance.
        """
        # Validate x.
        if not isinstance(x, ComplexAffineExpression):
            raise TypeError("Can only form the norm of an affine expression, "
                "not of {}.".format(type(x).__name__))

        complex = not isinstance(x, AffineExpression)

        if isinstance(p, tuple) and len(p) == 2:
            throw_deprecation_warning("Arguments 'p' and 'q' to Norm must be "
                "given separately in the future.", decoratorLevel=1)
            p, q = p

        if q is None:
            q = p

        # Load p as a limtied precision fraction.
        if p == float("inf"):
            pNum = p
            pDen = 1
            pStr = glyphs.infty()
        else:
            pNum, pDen, p, pStr = make_fraction(p, denominator_limit)

        # Load q as a limtied precision fraction.
        if q == float("inf"):
            qNum = q
            qDen = 1
            qStr = glyphs.infty()
        else:
            qNum, qDen, q, qStr = make_fraction(q, denominator_limit)

        # Validate that p and q are in the allowed range.
        if p == q:
            if complex and p < 1:
                raise NotImplementedError(
                    "Complex p-norm requires {}.".format(glyphs.ge("p", "1")))
            elif p <= 0:
                raise ValueError(
                    "p-norm requires {}.".format(glyphs.gt("p", "0")))
        elif p != q:
            if complex:
                raise NotImplementedError(
                    "(p,q)-norm is not supported for complex expressions.")
            elif p < 1 or q < 1:
                raise ValueError("(p,q)-norm requires {} and {}."
                    .format(glyphs.ge("p", "1"), glyphs.ge("q", "1")))

        # Build the string representations.
        if len(x) == 1:
            typeStr = "Modulus" if complex else "Absolute Value"
            symbStr = glyphs.abs(x.string)
        elif p == q:
            vec = 1 in x.shape
            if p == 1:
                typeStr = "Manhattan Norm"
                symbStr = glyphs.pnorm(x.string, "sum")
            elif p == 2:
                typeStr = "Euclidean Norm" if vec else "Frobenius Norm"
                symbStr = glyphs.norm(x.string)
            elif p == float("inf"):
                typeStr = "Maximum Norm"
                symbStr = glyphs.pnorm(x.string, "max")
            else:
                if p < 1:
                    typeStr = "Generalized p-Norm" if vec else \
                        "Entrywise Generalized p-Norm"
                else:
                    typeStr = "Vector p-Norm" if vec else "Entrywise p-Norm"
                symbStr = glyphs.pnorm(x.string, pStr)
        else:
            if p == 1 and q == float("inf"):
                typeStr = "Maximum Absolute Column Sum"
            else:
                typeStr = "Matrix (p,q)-Norm"
            symbStr = glyphs.pqnorm(x.string, pStr, qStr)

        if complex:
            typeStr = "Complex " + typeStr

        # Reduce the complex to the real case.
        if complex:
            if len(x) == 1:
                # From modulus to real Euclidean norm.
                x = x.real // x.imag

                p, pNum, pDen = 2.0, 2, 1
                q, qNum, qDen = 2.0, 2, 1
            else:
                # From complex vector/entrywise p-norm to real matrix p,q-norm.
                if x.shape[0] == 1:
                    x = x.real // x.imag
                elif x.shape[1] == 1:
                    x = x.T.real // x.T.imag
                else:
                    x = x.vec.T.real // x.vec.T.imag

                assert p == q
                p, pNum, pDen = 2.0, 2, 1

        # Reduce an entrywise p-norm to a vector p-norm and make all vector
        # norms refer to column vectors.
        if p == q and x.shape[1] != 1:
            x = x.vec

        assert isinstance(x, AffineExpression)
        assert float(pNum) / float(pDen) == p and float(qNum) / float(qDen) == q

        self._x     = x
        self._pNum  = pNum
        self._pDen  = pDen
        self._qNum  = qNum
        self._qDen  = qDen
        self._limit = denominator_limit

        Expression.__init__(self, typeStr, symbStr)

    # --------------------------------------------------------------------------
    # Abstract method implementations and method overridings, except _predict.
    # --------------------------------------------------------------------------

    @cached_unary_operator
    def _get_refined(self):
        if self._x.constant:
            return AffineExpression.from_constant(self.value, 1, self.string)
        else:
            return self

    Subtype = namedtuple("Subtype", ("xShape", "pNum", "pDen", "qNum", "qDen"))

    def _get_subtype(self):
        # NOTE: The xShape field refers to the internal (real and potentially
        #       vectorized) representation of the normed expression.
        return self.Subtype(
            self._x.shape, self._pNum, self._pDen, self._qNum, self._qDen)

    def _get_value(self):
        value = self._x._get_value()

        if len(value) == 1:
            return abs(value)

        value = cvx2np(value)

        p, q = self.p, self.q

        if p == q:
            value = numpy.linalg.norm(numpy.ravel(value), p)
        else:
            columns = value.shape[1]
            value = numpy.linalg.norm([
                numpy.linalg.norm(numpy.ravel(value[:, j]), p)
                for j in range(columns)], q)

        return cvxopt.matrix(value)

    @cached_unary_operator
    def _get_variables(self):
        return self._x.variables

    def _is_convex(self):
        return self.p >= 1 or len(self._x) == 1

    def _is_concave(self):
        return self.p <= 1 and len(self._x) != 1

    def _replace_variables(self, var_map):
        return self.__class__(
            self._x._replace_variables(var_map), self.p, self.q, self._limit)

    # --------------------------------------------------------------------------
    # Python special method implementations, except constraint-creating ones.
    # --------------------------------------------------------------------------

    @convert_operands(scalarRHS=True)
    @refine_operands()
    def __mul__(self, other):
        if isinstance(other, AffineExpression):
            if not other.constant:
                raise NotImplementedError("You may only multiply a nonconstant "
                    "PICOS norm with a constant term.")

            if other.value < 0:
                raise NotImplementedError("You may only multiply a nonconstant "
                    "PICOS norm with a nonnegative term.")

            norm = Norm(other*self._x, self.p, self.q, self._limit)
            norm._typeStr = "Scaled " + norm._typeStr
            norm._symbStr = glyphs.clever_mul(self.string, other.string)
            return norm
        else:
            return NotImplemented

    @convert_operands(scalarRHS=True)
    @refine_operands()
    def __rmul__(self, other):
        if isinstance(other, AffineExpression):
            norm = self.__mul__(other)
            # NOTE: __mul__ always creates a fresh expression.
            norm._symbStr = glyphs.clever_mul(other.string, self.string)
            return norm
        else:
            return NotImplemented

    @convert_operands(scalarRHS=True)
    def __pow__(self, other):
        if isinstance(other, AffineExpression):
            if not other.constant or other.value != 2:
                raise NotImplementedError("You may only take a PICOS norm to "
                    "the power of two.")

            p, q = self.p, self.q

            if p == q and p == 2:
                square = (self._x | self._x)
                # NOTE: AffineExpression.__or__ creates a fresh expression.
                square._symbStr = glyphs.squared(self.string)
                return square
            else:
                raise NotImplementedError(
                    "You may only square an Euclidean or Frobenius norm.")
        else:
            return NotImplemented

    # --------------------------------------------------------------------------
    # Properties and functions that describe the expression.
    # --------------------------------------------------------------------------

    @property
    def p(self):
        """The parameter :math:`p`.

        This is a limited precision version of the paramter used when the norm
        was constructed.
        """
        return float(self._pNum) / float(self._pDen)

    @property
    def pnum(self):
        """The limited precision fraction numerator of :math:`p`."""
        return self._pNum

    @property
    def pden(self):
        """The limited precision fraction denominator of :math:`p`."""
        return self._pDen

    @property
    def q(self):
        """The parameter :math:`q`.

        This is a limited precision version of the paramter used when the norm
        was constructed.
        """
        return float(self._qNum) / float(self._qDen)

    @property
    def qnum(self):
        """The limited precision fraction numerator of :math:`q`."""
        return self._qNum

    @property
    def qden(self):
        """The limited precision fraction denominator of :math:`q`."""
        return self._qDen

    # --------------------------------------------------------------------------
    # Methods and properties that return modified copies.
    # --------------------------------------------------------------------------

    @property
    def x(self):
        """Real expression whose norm equals that of the original expression."""
        return self._x

    # --------------------------------------------------------------------------
    # Constraint-creating operators, and _predict.
    # --------------------------------------------------------------------------

    @classmethod
    def _predict(cls, subtype, relation, other):
        assert isinstance(subtype, cls.Subtype)

        xShape, pNum, pDen, qNum, qDen = subtype
        xLen = xShape[0] * xShape[1]
        p = float(pNum) / float(pDen)
        q = float(qNum) / float(qDen)

        if relation == operator.__le__:
            if not (xLen == 1 or p >= 1):
                return NotImplemented  # Not convex.

            if issubclass(other.clstype, AffineExpression) \
            and other.subtype.dim == 1:
                if xLen == 1:
                    return AbsoluteValueConstraint.make_type()
                elif p == q == 2:
                    return SOCConstraint.make_type(argdim=xLen)
                elif p == q:
                    return VectorNormConstraint.make_type(
                        xLen, pNum, pDen, Constraint.LE)
                else:
                    return MatrixNormConstraint.make_type(
                        xShape, pNum, pDen, qNum, qDen)
        elif relation == operator.__ge__:
            if not (xLen != 1 and p <= 1):
                return NotImplemented  # Not concave.

            assert p == q

            if issubclass(other.clstype, AffineExpression) \
            and other.subtype.dim == 1:
                return VectorNormConstraint.make_type(
                    xLen, pNum, pDen, Constraint.GE)

        return NotImplemented

    @convert_operands(scalarRHS=True)
    @validate_prediction
    @refine_operands()
    def __le__(self, other):
        if not self.convex:
            raise TypeError("Cannot upper-bound a nonconvex generalized norm.")

        if isinstance(other, AffineExpression):
            if len(self._x) == 1:
                return AbsoluteValueConstraint(self._x, other)
            elif self.p == self.q == 2:
                # NOTE: The custom string is necessary in case the original x
                #       was complex; otherwise the string would be something
                #       like "‖vec([Re(xᵀ); Im(xᵀ)])‖ ≤ b".
                return SOCConstraint(self._x, other,
                    customString=glyphs.le(self.string, other.string))
            elif self.p == self.q:
                return VectorNormConstraint(self, Constraint.LE, other)
            else:
                return MatrixNormConstraint(self, other)
        else:
            return NotImplemented

    @convert_operands(scalarRHS=True)
    @validate_prediction
    @refine_operands()
    def __ge__(self, other):
        if not self.concave:
            raise TypeError("Cannot lower-bound a nonconcave norm.")

        assert self.p == self.q

        if isinstance(other, AffineExpression):
            return VectorNormConstraint(self, Constraint.GE, other)
        else:
            return NotImplemented


class SpectralNorm(Expression):
    r"""The spectral norm of a matrix.

    This class can represent the spectral norm of a matrix-affine expression
    (real- or complex valued). The spectral norm is convex, so we can form
    expressions of the form ``SpectralNorm(X) <= t`` which are typically
    reformulated as LMIs that can be handled by SDP solvers.

    :Definition:

    If the normed expression is a matrix :math:`X`, then its spectral norm is

        .. math::

            \|X\|_2 = \max \{  \|Xu\|_2 : \|u\| \leq  1\}
                    = \sqrt{\lambda_{\max}(XX^*)},

    where :math:`\lambda_{\max}(\cdot)` denotes the largest eigenvalue of
    a matrix, and :math:`X^*` denotes the adjoint matrix of :math:`X`
    (i.e., the transposed matrix :math:`X^T` if :math:`X` is real-valued).

    Special cases:

    -   If :math:`X` is scalar, then :math:`\|X\|_2` reduces to the the absolute
        value (or modulus) :math:`|X|`.
    -   If :math:`X` is scalar, then :math:`\|X\|_2` coincides with the
        Euclidean norm of :math:`X`.

    """

    @convert_and_refine_arguments("x")
    def __init__(self, x):
        """Construct a :class:`SpectralNorm`.

        :param x: The affine expression to take the norm of.
        :type x: ~picos.expressions.ComplexAffineExpression
        """
        # Validate x.
        if not isinstance(x, ComplexAffineExpression):
            raise TypeError("Can only form the spectral norm of an affine "
                "expression, not of {}.".format(type(x).__name__))

        complex = not isinstance(x, AffineExpression)

        # Build the string representations.
        if len(x) == 1:
            typeStr = "Modulus" if complex else "Absolute Value"
            symbStr = glyphs.abs(x.string)
        elif 1 in x.shape:
            typeStr = "Euclidean Norm"
            symbStr = glyphs.norm(x.string)
        else:
            typeStr = "Spectral Norm"
            symbStr = glyphs.spnorm(x.string)

        if complex:
            typeStr = "Complex " + typeStr

        self._x = x
        self._complex = complex
        Expression.__init__(self, typeStr, symbStr)

    # --------------------------------------------------------------------------
    # Abstract method implementations and method overridings, except _predict.
    # --------------------------------------------------------------------------

    @cached_unary_operator
    def _get_refined(self):
        if self._x.constant:
            return AffineExpression.from_constant(self.value, 1, self.string)
        elif len(self._x) == 1 or (1 in self._x.shape):
            return Norm(self._x)
        else:
            return self

    Subtype = namedtuple("Subtype", ("argshape", "complex", "hermitian"))

    def _get_subtype(self):
        return self.Subtype(self._x.shape, self._complex, self._x.hermitian)

    def _get_value(self):
        value = self._x._get_value()
        value = cvx2np(value)
        value = numpy.linalg.norm(value, 2)
        return cvxopt.matrix(value)

    @cached_unary_operator
    def _get_variables(self):
        return self._x.variables

    def _is_convex(self):
        return True

    def _is_concave(self):
        return False

    def _replace_variables(self, var_map):
        return self.__class__(self._x._replace_variables(var_map))

    # --------------------------------------------------------------------------
    # Python special method implementations, except constraint-creating ones.
    # --------------------------------------------------------------------------

    @convert_operands(scalarRHS=True)
    @refine_operands()
    def __mul__(self, other):
        if isinstance(other, AffineExpression):
            if not other.constant:
                raise NotImplementedError("You may only multiply a nonconstant "
                    "PICOS norm with a constant term.")

            if other.value < 0:
                raise NotImplementedError("You may only multiply a nonconstant "
                    "PICOS norm with a nonnegative term.")

            norm = SpectralNorm(other*self._x)
            norm._typeStr = "Scaled " + norm._typeStr
            norm._symbStr = glyphs.clever_mul(self.string, other.string)
            return norm
        else:
            return NotImplemented

    @convert_operands(scalarRHS=True)
    @refine_operands()
    def __rmul__(self, other):
        if isinstance(other, AffineExpression):
            norm = self.__mul__(other)
            # NOTE: __mul__ always creates a fresh expression.
            norm._symbStr = glyphs.clever_mul(other.string, self.string)
            return norm
        else:
            return NotImplemented

    # --------------------------------------------------------------------------
    # Methods and properties that return modified copies.
    # --------------------------------------------------------------------------

    @property
    def x(self):
        """Real expression whose norm equals that of the original expression."""
        return self._x

    # --------------------------------------------------------------------------
    # Constraint-creating operators, and _predict.
    # --------------------------------------------------------------------------

    @classmethod
    def _predict(cls, subtype, relation, other):
        assert isinstance(subtype, cls.Subtype)

        arg_shape, arg_complex, arg_hermitian = subtype
        xLen = arg_shape[0] * arg_shape[1]

        if relation == operator.__le__:
            if issubclass(other.clstype, AffineExpression) \
            and other.subtype.dim == 1:
                if xLen == 1:
                    return AbsoluteValueConstraint.make_type()
                elif 1 in arg_shape:
                    assert False, "Unexpected case (should have been refined)"
                else:
                    return SpectralNormConstraint.make_type(
                        arg_shape, arg_complex, arg_hermitian)
        elif relation == operator.__ge__:
            return NotImplemented  # Not concave.

        return NotImplemented

    @convert_operands(scalarRHS=True)
    @validate_prediction
    @refine_operands()
    def __le__(self, other):
        assert self.convex

        if isinstance(other, AffineExpression):
            if len(self._x) == 1:
                return AbsoluteValueConstraint(self._x, other)
            elif 1 in self._x.shape:
                assert False, "Unexpected case (should have been refined)"
            else:
                return SpectralNormConstraint(self, other)
        else:
            return NotImplemented


class NuclearNorm(Expression):
    r"""The nuclear norm of a matrix.

    This class can represent the nuclear norm of a matrix-affine expression
    (real- or complex valued). The nuclear norm is convex, so we can form
    expressions of the form ``NuclearNorm(X) <= t`` which are typically
    reformulated as LMIs that can be handled by SDP solvers.

    :Definition:

    If the normed expression is a matrix :math:`X`, then its nuclear norm is

        .. math::

            \|X\|_* =  \operatorname{trace}\ (X^*X)^{1/2}
                    = \sum_{i=1}^{\min(n,m)} \sigma_i(X)

    where the :math:`\sigma_i(X)` denote the singular values of
    a :math:`X`, and :math:`X^*` denotes the adjoint matrix of :math:`X`
    (i.e., the transposed matrix :math:`X^T` if :math:`X` is real-valued).

    Special cases:

    -   If :math:`X` is scalar, then :math:`\|X\|_*` reduces to the the absolute
        value (or modulus) :math:`|X|`.
    -   If :math:`X` is scalar, then :math:`\|X\|_*` coincides with the
        Euclidean norm of :math:`X`.

    """

    @convert_and_refine_arguments("x")
    def __init__(self, x):
        """Construct a :class:`NuclearNorm`.

        :param x: The affine expression to take the norm of.
        :type x: ~picos.expressions.ComplexAffineExpression
        """
        # Validate x.
        if not isinstance(x, ComplexAffineExpression):
            raise TypeError("Can only form the nuclear norm of an affine "
                "expression, not of {}.".format(type(x).__name__))

        complex = not isinstance(x, AffineExpression)

        # Build the string representations.
        if len(x) == 1:
            typeStr = "Modulus" if complex else "Absolute Value"
            symbStr = glyphs.abs(x.string)
        elif 1 in x.shape:
            typeStr = "Euclidean Norm"
            symbStr = glyphs.norm(x.string)
        else:
            typeStr = "Nuclear Norm"
            symbStr = glyphs.ncnorm(x.string)

        if complex:
            typeStr = "Complex " + typeStr

        self._x = x
        self._complex = complex
        Expression.__init__(self, typeStr, symbStr)

    # --------------------------------------------------------------------------
    # Abstract method implementations and method overridings, except _predict.
    # --------------------------------------------------------------------------

    @cached_unary_operator
    def _get_refined(self):
        if self._x.constant:
            return AffineExpression.from_constant(self.value, 1, self.string)
        elif len(self._x) == 1 or (1 in self._x.shape):
            return Norm(self._x)
        else:
            return self

    Subtype = namedtuple("Subtype", ("argshape", "complex", "hermitian"))

    def _get_subtype(self):
        return self.Subtype(self._x.shape, self._complex, self._x.hermitian)

    def _get_value(self):
        value = self._x._get_value()
        value = cvx2np(value)
        value = numpy.linalg.norm(value, 'nuc')
        return cvxopt.matrix(value)

    @cached_unary_operator
    def _get_variables(self):
        return self._x.variables

    def _is_convex(self):
        return True

    def _is_concave(self):
        return False

    def _replace_variables(self, var_map):
        return self.__class__(self._x._replace_variables(var_map))

    # --------------------------------------------------------------------------
    # Python special method implementations, except constraint-creating ones.
    # --------------------------------------------------------------------------

    @convert_operands(scalarRHS=True)
    @refine_operands()
    def __mul__(self, other):
        if isinstance(other, AffineExpression):
            if not other.constant:
                raise NotImplementedError("You may only multiply a nonconstant "
                    "PICOS norm with a constant term.")

            if other.value < 0:
                raise NotImplementedError("You may only multiply a nonconstant "
                    "PICOS norm with a nonnegative term.")

            norm = NuclearNorm(other*self._x)
            norm._typeStr = "Scaled " + norm._typeStr
            norm._symbStr = glyphs.clever_mul(self.string, other.string)
            return norm
        else:
            return NotImplemented

    @convert_operands(scalarRHS=True)
    @refine_operands()
    def __rmul__(self, other):
        if isinstance(other, AffineExpression):
            norm = self.__mul__(other)
            # NOTE: __mul__ always creates a fresh expression.
            norm._symbStr = glyphs.clever_mul(other.string, self.string)
            return norm
        else:
            return NotImplemented

    # --------------------------------------------------------------------------
    # Methods and properties that return modified copies.
    # --------------------------------------------------------------------------

    @property
    def x(self):
        """Real expression whose norm equals that of the original expression."""
        return self._x

    # --------------------------------------------------------------------------
    # Constraint-creating operators, and _predict.
    # --------------------------------------------------------------------------

    @classmethod
    def _predict(cls, subtype, relation, other):
        assert isinstance(subtype, cls.Subtype)

        arg_shape, arg_complex, arg_hermitian = subtype
        xLen = arg_shape[0] * arg_shape[1]

        if relation == operator.__le__:
            if issubclass(other.clstype, AffineExpression) \
            and other.subtype.dim == 1:
                if xLen == 1:
                    return AbsoluteValueConstraint.make_type()
                elif 1 in arg_shape:
                    assert False, "Unexpected case (should have been refined)"
                else:
                    return NuclearNormConstraint.make_type(
                               arg_shape, arg_complex, arg_hermitian)
        elif relation == operator.__ge__:
            return NotImplemented  # Not concave.

        return NotImplemented

    @convert_operands(scalarRHS=True)
    @validate_prediction
    @refine_operands()
    def __le__(self, other):
        assert self.convex

        if isinstance(other, AffineExpression):
            if len(self._x) == 1:
                return AbsoluteValueConstraint(self._x, other)
            elif 1 in self._x.shape:
                assert False, "Unexpected case (should have been refined)"
            else:
                return NuclearNormConstraint(self, other)
        else:
            return NotImplemented


# --------------------------------------
__all__ = api_end(_API_START, globals())
